#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pyqmri.models.template import BaseModel, constraints, DTYPE
import numexpr as ne
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
plt.ion()


class Model(BaseModel):
    """ Realization of the Inversion Recovery Look-Locker model.

    Attributes:
      TR (float): Pause in ms after last read out
      fa (float): Flip angle in radiants
      sin_phi (numpy.Array): Precompute sine of the B1 corrected flip angle
      cos_phi (numpy.Array): Precomputed co-sine of the B1 corrected flip angle
      guess (numpy.Array): Initial guess
      tau (float): Time petween each acquired spoke in ms
      td (float): Gradient delay in ms
      NLL (int): Number of Look-Locker excitations
      Nproj (int): Number of projections used for reconstruction
      Nproj_measured (int): Total number of acquired projections
    """
    def __init__(self, par):
        super().__init__(par)

        self.TR = par['time_per_slice'] - \
            (par['tau'] * par['Nproj_measured'] + par['gradient_delay'])
        self.fa = par["flip_angle(s)"] * np.pi / 180

        self.tau = par["tau"]
        self.td = par["gradient_delay"]
        self.NLL = par["NScan"]
        self.Nproj = par["Nproj"]
        self.Nproj_measured = par["Nproj_measured"]

        phi_corr = np.zeros_like(par["fa_corr"], dtype=DTYPE)
        phi_corr = np.real(
            self.fa) * np.real(par["fa_corr"]) + 1j *\
            np.imag(self.fa) * np.imag(par["fa_corr"])

        self.sin_phi = np.sin(phi_corr)
        self.cos_phi = np.cos(phi_corr)

        par["unknowns_TGV"] = 2
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"] + par["unknowns_H1"]

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        self.scale = 100

        self.constraints.append(constraints(0, 300, False))
        self.constraints.append(constraints(
            np.exp(-self.scale / 10) / self.uk_scale[1],
            np.exp(-self.scale / 5500) / self.uk_scale[1], True))

    def rescale(self, x):
        """ Rescales each unknown by its scaling factor.

          This realization takes into account that E1 instead of T1 is fitted.
          Each exponential containing T1 is substituted by E1 = exp(-scale/T1).
          Thus, rescaling of T1 is performed by -scale/log(E1).

        Args:
          x (numpy.Array): The array of unknowns to be rescaled
        Returns:
          numpy.Array: The rescaled array of unknowns
        """
        tmp_x = np.copy(x)
        tmp_x[0] *= self.uk_scale[0]
        tmp_x[1] = -self.scale / np.log(tmp_x[1] * self.uk_scale[1])
        return tmp_x

    def plot_unknowns(self, x, dim_2D=False):
        M0 = np.abs(x[0, ...] * self.uk_scale[0])
        T1 = np.abs(-self.scale / np.log(x[1, ...] * self.uk_scale[1]))
        M0_min = M0.min()
        M0_max = M0.max()
        T1_min = T1.min()
        T1_max = T1.max()
        if dim_2D:
            if not self.figure:
                plt.ion()
                self.figure, self.ax = plt.subplots(1, 2, figsize=(12, 5))
                self.M0_plot = self.ax[0].imshow((M0))
                self.ax[0].set_title('Proton Density in a.u.')
                self.ax[0].axis('off')
                self.figure.colorbar(self.M0_plot, ax=self.ax[0])
                self.T1_plot = self.ax[1].imshow((T1))
                self.ax[1].set_title('T1 in  ms')
                self.ax[1].axis('off')
                self.figure.colorbar(self.T1_plot, ax=self.ax[1])
                self.figure.tight_layout()
                plt.draw()
                plt.pause(1e-10)
            else:
                self.M0_plot.set_data((M0))
                self.M0_plot.set_clim([M0_min, M0_max])
                self.T1_plot.set_data((T1))
                self.T1_plot.set_clim([T1_min, T1_max])
                plt.draw()
                plt.pause(1e-10)
        else:
            [z, y, x] = M0.shape
            self.ax = []
            if not self.figure:
                plt.ion()
                self.figure = plt.figure(figsize=(12, 6))
                self.figure.subplots_adjust(hspace=0, wspace=0)
                self.gs = gridspec.GridSpec(2, 6, width_ratios=[
                                            x / (20 * z), x / z, 1,
                                            x / z, 1, x / (20 * z)],
                                            height_ratios=[x / z, 1])
                self.figure.tight_layout()
                self.figure.patch.set_facecolor(plt.cm.viridis.colors[0])
                for grid in self.gs:
                    self.ax.append(plt.subplot(grid))
                    self.ax[-1].axis('off')

                self.M0_plot = self.ax[1].imshow(
                    (M0[int(self.NSlice / 2), ...]))
                self.M0_plot_cor = self.ax[7].imshow(
                    (M0[:, int(M0.shape[1] / 2), ...]))
                self.M0_plot_sag = self.ax[2].imshow(
                    np.flip((M0[:, :, int(M0.shape[-1] / 2)]).T, 1))
                self.ax[1].set_title('Proton Density in a.u.', color='white')
                self.ax[1].set_anchor('SE')
                self.ax[2].set_anchor('SW')
                self.ax[7].set_anchor('NW')
                cax = plt.subplot(self.gs[:, 0])
                cbar = self.figure.colorbar(self.M0_plot, cax=cax)
                cbar.ax.tick_params(labelsize=12, colors='white')
                cax.yaxis.set_ticks_position('left')
                for spine in cbar.ax.spines:
                    cbar.ax.spines[spine].set_color('white')
                plt.draw()
                plt.pause(1e-10)

                self.T1_plot = self.ax[3].imshow(
                    (T1[int(self.NSlice / 2), ...]))
                self.T1_plot_cor = self.ax[9].imshow(
                    (T1[:, int(T1.shape[1] / 2), ...]))
                self.T1_plot_sag = self.ax[4].imshow(
                    np.flip((T1[:, :, int(T1.shape[-1] / 2)]).T, 1))
                self.ax[3].set_title('T1 in  ms', color='white')
                self.ax[3].set_anchor('SE')
                self.ax[4].set_anchor('SW')
                self.ax[9].set_anchor('NW')
                cax = plt.subplot(self.gs[:, 5])
                cbar = self.figure.colorbar(self.T1_plot, cax=cax)
                cbar.ax.tick_params(labelsize=12, colors='white')
                for spine in cbar.ax.spines:
                    cbar.ax.spines[spine].set_color('white')
                plt.draw()
                plt.pause(1e-10)
            else:
                self.M0_plot.set_data((M0[int(self.NSlice / 2), ...]))
                self.M0_plot_cor.set_data((M0[:, int(M0.shape[1] / 2), ...]))
                self.M0_plot_sag.set_data(
                    np.flip((M0[:, :, int(M0.shape[-1] / 2)]).T, 1))
                self.M0_plot.set_clim([M0_min, M0_max])
                self.M0_plot_cor.set_clim([M0_min, M0_max])
                self.M0_plot_sag.set_clim([M0_min, M0_max])
                self.T1_plot.set_data((T1[int(self.NSlice / 2), ...]))
                self.T1_plot_cor.set_data((T1[:, int(T1.shape[1] / 2), ...]))
                self.T1_plot_sag.set_data(
                    np.flip((T1[:, :, int(T1.shape[-1] / 2)]).T, 1))
                self.T1_plot.set_clim([T1_min, T1_max])
                self.T1_plot_sag.set_clim([T1_min, T1_max])
                self.T1_plot_cor.set_clim([T1_min, T1_max])
                plt.draw()
                plt.pause(1e-10)



    def _execute_forward_3D(self, x):
        S = np.zeros(
            (self.NLL,
             self.Nproj,
             self.NSlice,
             self.dimY,
             self.dimX),
            dtype=DTYPE)

        TR = self.TR
        tau = self.tau
        td = self.td
        sin_phi = self.sin_phi
        cos_phi = self.cos_phi
        N = self.Nproj_measured
        scale = self.scale
        Efit = x[1, ...] * self.uk_scale[1]
        Etau = Efit**(tau / scale)
        Etr = Efit**(TR / scale)
        Etd = Efit**(td / scale)
        M0 = x[0, ...]
        M0_sc = self.uk_scale[0]

        F = (1 - Etau) / (1 - Etau * cos_phi)
        Q = (-Etr * Etd * F * (-(Etau * cos_phi)**(N - 1) + 1) *
             cos_phi + Etr *
             Etd - 2 * Etd + 1) / (Etr * Etd * (Etau * cos_phi)**(N - 1) *
                                   cos_phi + 1)
        Q_F = Q - F

        def numexpeval_S(M0, M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau):
            return ne.evaluate(
                "M0*M0_sc*((Etau*cos_phi)**(n - 1)*Q_F + F)*sin_phi")
        for i in range(self.NLL):
            for j in range(self.Nproj):
                n = i * self.Nproj + j + 1
                S[i, j, ...] = numexpeval_S(
                    M0, M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau)

        return np.array(np.mean(S, axis=1, dtype=DTYPE), dtype=DTYPE)

    def _execute_gradient_3D(self, x):
        grad = np.zeros(
            (2,
             self.NLL,
             self.Nproj,
             self.NSlice,
             self.dimY,
             self.dimX),
            dtype=DTYPE)
        M0_sc = self.uk_scale[0]
        TR = self.TR
        tau = self.tau
        td = self.td
        sin_phi = self.sin_phi
        cos_phi = self.cos_phi
        N = self.Nproj_measured
        scale = self.scale
        Efit = x[1, ...] * self.uk_scale[1]
        Etau = Efit**(tau / scale)
        Etr = Efit**(TR / scale)
        Etd = Efit**(td / scale)
        M0 = x[0, ...]
        M0_sc = self.uk_scale[0]

        F = (1 - Etau) / (1 - Etau * cos_phi)
        Q = (-Etr * Etd * (-Etau + 1) * (-(Etau * cos_phi)**(N - 1) + 1) *
             cos_phi / (-Etau * cos_phi + 1) + Etr * Etd - 2 * Etd + 1) / \
            (Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi + 1)
        Q_F = Q - F
        tmp1 = ((-TR * Etr * Etd * (-Etau + 1) *
                 (-(Etau * cos_phi)**(N - 1) + 1)
                 * cos_phi / (x[1, ...] * scale * (-Etau * cos_phi + 1)) +
                 TR * Etr * Etd / (x[1, ...] * scale) - tau * Etr * Etau *
                 Etd * (-Etau + 1) *
                 (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi**2 / (x[1, ...] *
                 scale * (-Etau * cos_phi + 1)**2) +
                 tau * Etr * Etau * Etd * (-(Etau * cos_phi)**(N - 1) + 1) *
                 cos_phi / (x[1, ...] * scale * (-Etau * cos_phi + 1)) +
                 tau * Etr * Etd * (Etau * cos_phi)**(N - 1) *
                 (N - 1) * (-Etau + 1) * cos_phi / (x[1, ...] * scale *
                 (-Etau * cos_phi + 1)) -
                 td * Etr * Etd * (-Etau + 1) *
                 (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi / (x[1, ...] *
                 scale * (-Etau * cos_phi + 1)) +
                 td * Etr * Etd / (x[1, ...] * scale) - 2 *
                 td * Etd / (x[1, ...] * scale)) /
                (Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi + 1) +
                (-TR * Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi /
                 (x[1, ...] * scale) - tau * Etr * Etd *
                 (Etau * cos_phi)**(N - 1) * (N - 1) *
                 cos_phi / (x[1, ...] * scale) - td * Etr * Etd *
                 (Etau * cos_phi)**(N - 1) * cos_phi / (x[1, ...] *
                 scale)) * (-Etr * Etd * (-Etau + 1) *
                (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi /
                 (-Etau * cos_phi + 1) + Etr * Etd - 2 * Etd + 1) /
                (Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi + 1)**2 -
                tau * Etau * (-Etau + 1) * cos_phi / (x[1, ...] * scale *
                (-Etau * cos_phi + 1)**2)
                + tau * Etau / (x[1, ...] * scale * (-Etau * cos_phi + 1)))

        tmp2 = tau * Etau * (-Etau + 1) * cos_phi / (x[1, ...] * scale * (
            -Etau * cos_phi + 1)**2) - tau * Etau / (x[1, ...] * scale * (
                -Etau * cos_phi + 1))

        tmp3 = (-(-Etau + 1) / (-Etau * cos_phi + 1) + (-Etr * Etd *
                (-Etau + 1) * (-(Etau * cos_phi)**(N - 1) + 1) * cos_phi /
                (-Etau * cos_phi + 1) + Etr * Etd - 2 * Etd + 1) / (
                    Etr * Etd * (Etau * cos_phi)**(N - 1) * cos_phi +
                    1)) / (x[1, ...] * scale)

        def numexpeval_M0(M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau):
            return ne.evaluate(
                "M0_sc*((Etau*cos_phi)**(n - 1)*Q_F + F)*sin_phi")

        def numexpeval_T1(
                M0,
                M0_sc,
                Etau,
                cos_phi,
                sin_phi,
                n,
                tmp1,
                tmp2,
                tmp3,
                tau):
            return ne.evaluate(
                "M0*M0_sc*((Etau*cos_phi)**(n - 1)*tmp1 + "
                "tmp2 + tau*(Etau*cos_phi)**(n - 1)*(n - 1)*tmp3)*sin_phi")

        for i in range(self.NLL):
            for j in range(self.Nproj):
                n = i * self.Nproj + j + 1

                grad[0, i, j, ...] = numexpeval_M0(
                    M0_sc, sin_phi, cos_phi, n, Q_F, F, Etau)
                grad[1, i, j, ...] = numexpeval_T1(
                    M0, M0_sc, Etau, cos_phi, sin_phi, n,
                    tmp1, tmp2, tmp3, tau)

        return np.array(np.mean(grad, axis=2, dtype=DTYPE), dtype=DTYPE)

    def computeInitialGuess(self, *args):

        test_T1 = 800 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_M0 = np.ones((self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_T1 = np.exp(-self.scale / (test_T1))

        x = np.array([test_M0 / self.uk_scale[0],
                      test_T1 / self.uk_scale[1]], dtype=DTYPE)
        self.guess = x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 30 10:20:02 2019

@author: omaier
"""

import pyopencl as cl


class CLProgram(object):
    def __init__(self, ctx, code):
        self._cl_prg = cl.Program(ctx, code)
        self._cl_prg.build("-cl-mad-enable -cl-fast-relaxed-math")
        self._cl_kernels = self._cl_prg.all_kernels()
        for kernel in self._cl_kernels:
            self.__dict__[kernel.function_name] = kernel



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" This module holds the classes for different FFT operators.
"""
import numpy as np
import pyopencl as cl
import pyopencl.array as clarray
from gpyfft.fft import FFT
from pkg_resources import resource_filename
from pyqmri._helper_fun._calckbkernel import calckbkernel
from pyqmri._helper_fun import CLProgram as Program


class PyOpenCLFFT():
    """ Base class for FFT calculation.

    This class serves as the base class for all FFT object used in
    the varous optimization algorithms. It provides a factory method
    to generate a FFT object based on the input.

    Attributes:
      DTYPE (Numpy.Type):
        The comlex precission type. Currently complex64 is used.
      DTYPE_real (Numpy.Type):
        The real precission type. Currently float32 is used.
      ctx (PyOpenCL.Context):
        The context for the PyOpenCL computations.
      queue (PyOpenCL.Queue):
        The computation Queue for the PyOpenCL kernels.
    """
    def __init__(self, ctx, queue, DTYPE, DTYPE_real):
        """Initiate the core components of the FFT object.

        Args:
          ctx (PyOpenCL.Context):
            The context for the PyOpenCL computations.
          queue (PyOpenCL.Queue):
            The computation Queue for the PyOpenCL kernels.
          DTYPE (Numpy.Type):
            The comlex precission type. Currently complex64 is used.
          DTYPE_real (Numpy.Type):
            The real precission type. Currently float32 is used.
        """
        self.DTYPE = DTYPE
        self.DTYPE_real = DTYPE_real
        self.ctx = ctx
        self.queue = queue

    @staticmethod
    def create(ctx,
               queue,
               par,
               kwidth=5,
               fft_dim=(
                   1,
                   2),
               klength=200,
               DTYPE=np.complex64,
               DTYPE_real=np.float32,
               radial=False,
               SMS=False,
               streamed=False):
        """The FFT factory method.
        Based on the inputs this method decides which FFT object should be
        returned.

        Args:
          ctx (PyOpenCL.Context):
            The context for the PyOpenCL computations.
          queue (PyOpenCL.Queue):
            The computation Queue for the PyOpenCL kernels.
          par (dict):
            A python dict containing the necessary information to setup the
            object. Needs to contain the number of slices (NSlice), number of
            scans (NScan), image dimensions (dimX, dimY), number of coils (NC),
            sampling points (N) and read outs (NProj) a PyOpenCL queue (queue)
            and the complex coil sensitivities (C).
          kwidth (int):
            The width of the sampling kernel for regridding of non-uniform
            kspace samples.
          fft_dim (tuple of ints):
            The dimensions which should be transformed. Defaults to 1 and 2
            corresponding the the last two of fft_dim.
          klength (int):
            The length of the kernel lookup table which samples the contineous
            gridding kernel.
          DTYPE (Numpy.Type):
            The comlex precission type. Currently complex64 is used.
          DTYPE_real (Numpy.Type):
            The real precission type. Currently float32 is used.
          radial (bool):
            Switch for Cartesian (False) and non-Cartesian (True) FFT.
          SMS (bool):
            Switch between Simultaneous Multi Slice reconstruction (True) and
            simple slice by slice reconstruction.
          streamed (bool):
            Switch between normal reconstruction in one big block versus
            streamed reconstruction of smaller blocks.

        Returns:
          PyOpenCLFFT object:
            The setup FFT object.

        Raises:
          AssertionError:
            If the Combination of passed flags to choose the
            FFT aren't compatible with each other. E.g.: Radial and SMS True.
        """
        if not streamed:
            if radial is True and SMS is False:
                obj = PyOpenCLRadialNUFFT(
                    ctx,
                    queue,
                    par,
                    kwidth=kwidth,
                    fft_dim=fft_dim,
                    klength=klength,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
            elif SMS is True and radial is False:
                raise AssertionError("SMS not implemented")
            elif SMS is False and radial is False:
                if "field_map" in par.keys():
                    obj = PyOpenCLFieldMapNUFFT(
                        ctx,
                        queue,
                        par,
                        fft_dim=fft_dim,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
                else:
                    obj = PyOpenCLCartNUFFT(
                        ctx,
                        queue,
                        par,
                        fft_dim=fft_dim,
                        DTYPE=DTYPE,
                        DTYPE_real=DTYPE_real)
            else:
                raise AssertionError("Combination of Radial "
                                     "and SMS not allowed")
            if DTYPE == np.complex128:
                print('Using double precission')
                file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_gridding_double.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
            else:
                print('Using single precission')
                file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_gridding_single.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
        else:
            if radial is True and SMS is False:
                obj = PyOpenCLRadialNUFFTStreamed(
                    ctx,
                    queue,
                    par,
                    kwidth=kwidth,
                    fft_dim=fft_dim,
                    klength=klength,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
            elif SMS is True and radial is False:
                raise AssertionError("SMS not implemented")
            elif SMS is False and radial is False:
                obj = PyOpenCLCartNUFFTStreamed(
                    ctx,
                    queue,
                    par,
                    fft_dim=fft_dim,
                    DTYPE=DTYPE,
                    DTYPE_real=DTYPE_real)
            else:
                raise AssertionError("Combination of Radial "
                                     "and SMS not allowed")
            if DTYPE == np.complex128:
                print('Using double precission')
                file = open(
                    resource_filename(
                        'pyqmri',
                        'kernels/OpenCL_gridding_slicefirst_double.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
            else:
                print('Using single precission')
                file = open(
                    resource_filename(
                        'pyqmri',
                        'kernels/OpenCL_gridding_slicefirst_single.c'))
                obj.prg = Program(
                    obj.ctx,
                    file.read())
        file.close()
        return obj


class PyOpenCLRadialNUFFT(PyOpenCLFFT):
    """ Non-uniform FFT object

    This class performs the non-uniform FFT (NUFFT) operation. Linear
    interpolation of a sampled gridding kernel is used to regrid points
    from the non-cartesian grid back on the cartesian grid.

    Attributes:
      traj (PyOpenCL.Array):
        The comlex sampling trajectory
      dcf (PyOpenCL.Array):
        The densitiy compenation function
      ogf (float):
        The overgriddingfactor for non-cartesian k-spaces.
      fft_shape (tuple of ints):
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale (float32):
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      cl_kerneltable (PyOpenCL.Buffer):
        The gridding lookup table as read only Buffer
      cl_deapo (PyOpenCL.Buffer):
        The deapodization lookup table as read only Buffer
      par_fft (int):
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft (gpyfft.fft.FFT):
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      prg (PyOpenCL.Program):
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """
    def __init__(
            self,
            ctx,
            queue,
            par,
            kwidth=5,
            fft_dim=(
                1,
                2),
            klength=200,
            DTYPE=np.complex64,
            DTYPE_real=np.float32):
        """ Setup the NUFFT object
        Args:
          ctx (PyOpenCL.Context):
            The context for the PyOpenCL computations.
          queue (PyOpenCL.Queue):
            The computation Queue for the PyOpenCL kernels.
          par (dict): A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          kwidth (int):
            The width of the sampling kernel for regridding of non-uniform
            kspace samples.
          klength (int):
            The length of the kernel lookup table which samples the contineous
            gridding kernel.
          fft_dim (tuple of ints):
            The dimensions which should be transformed. Defaults
            to 1 and 2 corresponding the the last two of fft_dim.
          DTYPE (Numpy.Type):
            The comlex precission type. Currently complex64 is used.
          DTYPE_real (Numpy.Type):
            The real precission type. Currently float32 is used.
        """
        super().__init__(ctx, queue, DTYPE, DTYPE_real)
        self.ogf = par["N"]/par["dimX"]
        self.fft_shape = (
            par["NScan"] *
            par["NC"] *
            par["NSlice"],
            int(par["dimY"]*self.ogf),
            int(par["dimX"]*self.ogf))
        self.fft_scale = DTYPE_real(
            np.sqrt(np.prod(self.fft_shape[fft_dim[0]:])))

        (kerneltable, kerneltable_FT, u) = calckbkernel(
            kwidth, self.ogf, par["N"], klength)

        deapo = 1 / kerneltable_FT.astype(DTYPE_real)

        self.cl_kerneltable = cl.Buffer(
            self.ctx,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=kerneltable.astype(DTYPE_real).data)
        self.cl_deapo = cl.Buffer(
            self.ctx,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=deapo.data)
        self.dcf = clarray.to_device(self.queue, par["dcf"])
        self.traj = clarray.to_device(self.queue, par["traj"])
        self._tmp_fft_array = (
            clarray.empty(
                self.queue,
                (self.fft_shape),
                dtype=DTYPE))
        self.par_fft = int(self.fft_shape[0] / par["NScan"])
        self.fft = FFT(ctx, queue, self._tmp_fft_array[
            0:self.par_fft, ...],
                       out_array=self._tmp_fft_array[
                           0:self.par_fft, ...],
                       axes=fft_dim)

        self._kernelpoints = kerneltable.size
        self._kwidth = kwidth / 2
        self._check = np.ones(par["N"], dtype=DTYPE_real)
        self._check[1::2] = -1
        self._check = clarray.to_device(self.queue, self._check)
        self._gridsize = par["N"]

    def __del__(self):
        del self.traj
        del self.dcf
        del self._tmp_fft_array
        del self.cl_kerneltable
        del self.cl_deapo
        del self._check
        del self.queue
        del self.ctx
        del self.prg
        del self.fft

    def FFTH(self, sg, s, wait_for=[]):
        """ Perform the inverse (adjoint) NUFFT operation
        Args:
          sg (PyOpenCL.Array):
            The complex image data.
          s (PyOpenCL.Array):
            The non-uniformly gridded k-space
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=s.events +
                sg.events +
                self._tmp_fft_array.events +
                wait_for))
        # Grid k-space
        self._tmp_fft_array.add_event(
            self.prg.grid_lut(
                self.queue,
                (s.shape[0], s.shape[1] * s.shape[2],
                 s.shape[-2] * self._gridsize),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.traj.data,
                np.int32(self._gridsize),
                self.DTYPE_real(self._kwidth / self._gridsize),
                self.dcf.data,
                self.cl_kerneltable,
                np.int32(self._kernelpoints),
                wait_for=(wait_for + sg.events +
                          s.events + self._tmp_fft_array.events)))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        for j in range(s.shape[0]):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        return self.prg.deapo_adj(
            self.queue,
            (sg.shape[0] * sg.shape[1] *
             sg.shape[2], sg.shape[3], sg.shape[4]),
            None,
            sg.data,
            self._tmp_fft_array.data,
            self.cl_deapo,
            np.int32(self._tmp_fft_array.shape[-1]),
            self.DTYPE_real(self.fft_scale),
            self.DTYPE_real(self.ogf),
            wait_for=(wait_for + sg.events + s.events +
                      self._tmp_fft_array.events))

    def FFT(self, s, sg, wait_for=[]):
        """ Perform the forward NUFFT operation
        Args:
          s (PyOpenCL.Array):
            The non-uniformly gridded k-space.
          sg (PyOpenCL.Array):
            The complex image data.
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=s.events +
                sg.events +
                self._tmp_fft_array.events +
                wait_for))
        # Deapodization and Scaling
        self._tmp_fft_array.add_event(
            self.prg.deapo_fwd(
                self.queue,
                (sg.shape[0] * sg.shape[1] * sg.shape[2],
                 sg.shape[3], sg.shape[4]),
                None,
                self._tmp_fft_array.data,
                sg.data,
                self.cl_deapo,
                np.int32(self._tmp_fft_array.shape[-1]),
                self.DTYPE_real(1 / self.fft_scale),
                self.DTYPE_real(self.ogf),
                wait_for=wait_for + sg.events + self._tmp_fft_array.events))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        for j in range(s.shape[0]):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        # Resample on Spoke
        return self.prg.invgrid_lut(
            self.queue,
            (s.shape[0], s.shape[1] * s.shape[2], s.shape[-2] *
             self._gridsize),
            None,
            s.data,
            self._tmp_fft_array.data,
            self.traj.data,
            np.int32(self._gridsize),
            self.DTYPE_real(self._kwidth / self._gridsize),
            self.dcf.data,
            self.cl_kerneltable,
            np.int32(self._kernelpoints),
            wait_for=s.events + wait_for + self._tmp_fft_array.events)


class PyOpenCLCartNUFFT(PyOpenCLFFT):
    """ Cartesian FFT object

    This class performs the FFT operation.

    Attributes:
      fft_shape (tuple of ints):
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale (float32):
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      par_fft (int):
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft (gpyfft.fft.FFT):
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      mask (PyOpenCL.Array):
        The undersampling mask for the Cartesian grid.
      prg (PyOpenCL.Program):
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """
    def __init__(
            self,
            ctx,
            queue,
            par,
            fft_dim=(
                1,
                2),
            DTYPE=np.complex64,
            DTYPE_real=np.float32):
        """ Setup the FFT object
        Args:
          ctx (PyOpenCL.Context):
            The context for the PyOpenCL computations.
          queue (PyOpenCL.Queue):
            The computation Queue for the PyOpenCL kernels.
          par (dict): A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          fft_dim (tuple of ints):
            The dimensions which should be transformed. Defaults
            to 1 and 2 corresponding the the last two of fft_dim.
          DTYPE (Numpy.Type):
            The comlex precission type. Currently complex64 is used.
          DTYPE_real (Numpy.Type):
            The real precission type. Currently float32 is used.
        """
        super().__init__(ctx, queue, DTYPE, DTYPE_real)
        self.fft_shape = (
            par["NScan"] *
            par["NC"] *
            par["NSlice"],
            par["dimY"],
            par["dimX"])
        self.fft_scale = DTYPE_real(
            np.sqrt(np.prod(self.fft_shape[fft_dim[0]:])))
        self._tmp_fft_array = (
            clarray.empty(
                self.queue,
                self.fft_shape,
                dtype=DTYPE))
        self.par_fft = int(self.fft_shape[0] / par["NScan"])
        self.mask = clarray.to_device(self.queue, par["mask"])
        self.fft = FFT(ctx, queue, self._tmp_fft_array[
            0:self.par_fft, ...],
                       out_array=self._tmp_fft_array[
                           0:self.par_fft, ...],
                       axes=fft_dim)

    def __del__(self):
        del self._tmp_fft_array
        del self.queue
        del self.ctx
        del self.prg
        del self.mask
        del self.fft

    def FFTH(self, sg, s, wait_for=[]):
        """ Perform the inverse (adjoint) FFT operation
        Args:
          sg (PyOpenCL.Array):
            The complex image data.
          s (PyOpenCL.Array):
            The uniformly gridded k-space
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._tmp_fft_array.add_event(
            self.prg.maskingcpy(
                self.queue,
                (self._tmp_fft_array.shape),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.mask.data,
                wait_for=s.events))
        for j in range(s.shape[0]):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        return (
            self.prg.copy(
                self.queue,
                (sg.size,
                 ),
                None,
                sg.data,
                self._tmp_fft_array.data,
                self.DTYPE_real(
                    self.fft_scale)))

    def FFT(self, s, sg, wait_for=[]):
        """ Perform the forward FFT operation
        Args:
          s (PyOpenCL.Array):
            The uniformly gridded k-space.
          sg (PyOpenCL.Array):
            The complex image data.
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._tmp_fft_array.add_event(
            self.prg.copy(
                self.queue,
                (sg.size,
                 ),
                None,
                self._tmp_fft_array.data,
                sg.data,
                self.DTYPE_real(
                    1 /
                    self.fft_scale)))

        for j in range(s.shape[0]):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])
        return (
            self.prg.maskingcpy(
                self.queue,
                (self._tmp_fft_array.shape),
                None,
                s.data,
                self._tmp_fft_array.data,
                self.mask.data,
                wait_for=s.events+self._tmp_fft_array.events))


class PyOpenCLFieldMapNUFFT(PyOpenCLFFT):
    """ Cartesian FFT object

    This class performs the FFT operation.

    Attributes:
      fft_shape (tuple of ints):
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale (float32):
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      par_fft (int):
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft (gpyfft.fft.FFT):
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      mask (PyOpenCL.Array):
        The undersampling mask for the Cartesian grid.
      prg (PyOpenCL.Program):
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """
    def __init__(
            self,
            ctx,
            queue,
            par,
            fft_dim=(
                1,
                2),
            DTYPE=np.complex64,
            DTYPE_real=np.float32):
        """ Setup the FFT object
        Args:
          ctx (PyOpenCL.Context):
            The context for the PyOpenCL computations.
          queue (PyOpenCL.Queue):
            The computation Queue for the PyOpenCL kernels.
          par (dict): A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          fft_dim (tuple of ints):
            The dimensions which should be transformed. Defaults
            to 1 and 2 corresponding the the last two of fft_dim.
          DTYPE (Numpy.Type):
            The comlex precission type. Currently complex64 is used.
          DTYPE_real (Numpy.Type):
            The real precission type. Currently float32 is used.
        """
        super().__init__(ctx, queue, DTYPE, DTYPE_real)
        self.fft_shape = (
            par["NScan"] *
            par["NC"] *
            par["NSlice"],
            par["dimY"],
            par["dimX"])
        self._tmp_fft_array = (
            clarray.empty(
                self.queue,
                self.fft_shape,
                dtype=DTYPE))
        self.mask = clarray.to_device(self.queue, par["mask"])
        self.phase_map = clarray.to_device(self.queue,
                                           par["phase_map"])

    def __del__(self):
        del self._tmp_fft_array
        del self.queue
        del self.ctx
        del self.prg
        del self.mask

    def FFTH(self, sg, s, wait_for=[]):
        """ Perform the inverse (adjoint) FFT operation
        Args:
          sg (PyOpenCL.Array):
            The complex image data.
          s (PyOpenCL.Array):
            The uniformly gridded k-space
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._tmp_fft_array.add_event(
            self.prg.maskingcpy(
                self.queue,
                (self._tmp_fft_array.shape),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.mask.data,
                wait_for=s.events))

        return self.prg.addfieldadj(
                            self.queue,
                            sg.shape[-3:],
                            None,
                            sg.data,
                            self._tmp_fft_array.data,
                            self.phase_map.data,
                            np.int32(sg.shape[1]),
                            np.int32(sg.shape[0]),
                            wait_for=self._tmp_fft_array.events+sg.events)

    def FFT(self, s, sg, wait_for=[]):
        """ Perform the forward FFT operation
        Args:
          s (PyOpenCL.Array):
            The uniformly gridded k-space.
          sg (PyOpenCL.Array):
            The complex image data.
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """

        self._tmp_fft_array.add_event(self.prg.addfield(
                            self.queue,
                            sg.shape[-3:],
                            None,
                            self._tmp_fft_array.data,
                            sg.data,
                            self.phase_map.data,
                            np.int32(sg.shape[1]),
                            np.int32(sg.shape[0]),
                            wait_for=self._tmp_fft_array.events+sg.events))

        return (
            self.prg.maskingcpy(
                self.queue,
                (self._tmp_fft_array.shape),
                None,
                s.data,
                self._tmp_fft_array.data,
                self.mask.data,
                wait_for=s.events+self._tmp_fft_array.events))


class PyOpenCLRadialNUFFTStreamed(PyOpenCLFFT):
    """ The streamed version of the non-uniform FFT object

    This class performs the non-uniform FFT (NUFFT) operation. Linear
    interpolation of a sampled gridding kernel is used to regrid points
    from the non-cartesian grid back on the cartesian grid.

    Attributes:
      traj (PyOpenCL.Array):
        The comlex sampling trajectory
      dcf (PyOpenCL.Array):
        The densitiy compenation function
      ogf (float):
        The overgriddingfactor for non-cartesian k-spaces.
      fft_shape (tuple of ints):
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale (float32):
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      cl_kerneltable (PyOpenCL.Buffer):
        The gridding lookup table as read only Buffer
      cl_deapo (PyOpenCL.Buffer):
        The deapodization lookup table as read only Buffer
      par_fft (int):
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft (gpyfft.fft.FFT):
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      prg (PyOpenCL.Program):
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """
    def __init__(
            self,
            ctx,
            queue,
            par,
            kwidth=5,
            fft_dim=(
                1,
                2),
            klength=200,
            DTYPE=np.complex64,
            DTYPE_real=np.float32):

        super().__init__(ctx, queue, DTYPE, DTYPE_real)
        self.ogf = par["N"]/par["dimX"]
        self.fft_shape = (par["NScan"] *
                          par["NC"] *
                          (par["par_slices"] +
                           par["overlap"]),
                          int(par["dimY"]*self.ogf),
                          int(par["dimX"]*self.ogf))
        (kerneltable, kerneltable_FT, u) = calckbkernel(
            kwidth, self.ogf, par["N"], klength)
        self._kernelpoints = kerneltable.size

        self.fft_scale = DTYPE_real(self.fft_shape[-1])
        self.deapo = 1 / kerneltable_FT.astype(DTYPE_real)
        self._kwidth = kwidth / 2
        self.cl_kerneltable = cl.Buffer(
            self.queue.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=kerneltable.astype(DTYPE_real).data)
        self.cl_deapo = cl.Buffer(
            self.queue.context,
            cl.mem_flags.READ_ONLY | cl.mem_flags.COPY_HOST_PTR,
            hostbuf=self.deapo.data)
        self.dcf = clarray.to_device(self.queue, par["dcf"])
        self.traj = clarray.to_device(self.queue, par["traj"])
        self._tmp_fft_array = (
            clarray.empty(
                self.queue,
                self.fft_shape,
                dtype=DTYPE))
        self.par_fft = int(self.fft_shape[0] / par["NScan"])
        self.fft = FFT(ctx, queue, self._tmp_fft_array[
            0:self.par_fft, ...],
                       out_array=self._tmp_fft_array[
                           0:self.par_fft, ...],
                       axes=fft_dim)

        self._kernelpoints = kerneltable.size
        self._kwidth = kwidth / 2
        self._check = np.ones(par["N"], dtype=DTYPE_real)
        self._check[1::2] = -1
        self._check = clarray.to_device(self.queue, self._check)
        self._gridsize = par["N"]

    def __del__(self):
        del self.traj
        del self.dcf
        del self._tmp_fft_array
        del self.cl_kerneltable
        del self.cl_deapo
        del self._check
        del self.queue
        del self.ctx
        del self.prg
        del self.fft

    def FFTH(self, sg, s, wait_for=[]):
        """ Perform the inverse (adjoint) NUFFT operation
        Args:
          sg (PyOpenCL.Array):
            The complex image data.
          s (PyOpenCL.Array):
            The non-uniformly gridded k-space
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=self._tmp_fft_array.events +
                wait_for))
        # Grid k-space
        self._tmp_fft_array.add_event(
            self.prg.grid_lut(
                self.queue,
                (s.shape[0], s.shape[1] * s.shape[2],
                 s.shape[-2] * self._gridsize),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.traj.data,
                np.int32(self._gridsize),
                np.int32(sg.shape[2]),
                self.DTYPE_real(self._kwidth / self._gridsize),
                self.dcf.data,
                self.cl_kerneltable,
                np.int32(self._kernelpoints),
                wait_for=(wait_for + sg.events + s.events +
                          self._tmp_fft_array.events)))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        for j in range(int(self.fft_shape[0] / self.par_fft)):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:
                        (j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:
                        (j + 1) * self.par_fft, ...],
                    forward=False)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        # Deapodization and Scaling
        return self.prg.deapo_adj(self.queue,
                                  (sg.shape[0] * sg.shape[1] * sg.shape[2],
                                   sg.shape[3], sg.shape[4]),
                                  None,
                                  sg.data,
                                  self._tmp_fft_array.data,
                                  self.cl_deapo,
                                  np.int32(self._tmp_fft_array.shape[-1]),
                                  self.DTYPE_real(self.fft_scale),
                                  self.DTYPE_real(self.ogf),
                                  wait_for=(wait_for + sg.events + s.events +
                                            self._tmp_fft_array.events))

    def FFT(self, s, sg, wait_for=[]):
        """ Perform the forward NUFFT operation
        Args:
          s (PyOpenCL.Array):
            The non-uniformly gridded k-space.
          sg (PyOpenCL.Array):
            The complex image data.
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        # Zero tmp arrays
        self._tmp_fft_array.add_event(
            self.prg.zero_tmp(
                self.queue,
                (self._tmp_fft_array.size,
                 ),
                None,
                self._tmp_fft_array.data,
                wait_for=self._tmp_fft_array.events +
                wait_for))
        # Deapodization and Scaling
        self._tmp_fft_array.add_event(
            self.prg.deapo_fwd(
                self.queue,
                (sg.shape[0] * sg.shape[1] * sg.shape[2], sg.shape[3],
                 sg.shape[4]),
                None,
                self._tmp_fft_array.data,
                sg.data,
                self.cl_deapo,
                np.int32(self._tmp_fft_array.shape[-1]),
                self.DTYPE_real(1 / self.fft_scale),
                self.DTYPE_real(self.ogf),
                wait_for=wait_for + sg.events + self._tmp_fft_array.events))
        # FFT
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        for j in range(int(self.fft_shape[0] / self.par_fft)):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:
                        (j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:
                        (j + 1) * self.par_fft, ...],
                    forward=True)[0])
        self._tmp_fft_array.add_event(
            self.prg.fftshift(
                self.queue,
                (self.fft_shape[0],
                 self.fft_shape[1],
                 self.fft_shape[2]),
                None,
                self._tmp_fft_array.data,
                self._check.data))
        # Resample on Spoke

        return self.prg.invgrid_lut(
            self.queue,
            (s.shape[0], s.shape[1] * s.shape[2],
             s.shape[-2] * self._gridsize),
            None,
            s.data,
            self._tmp_fft_array.data,
            self.traj.data,
            np.int32(self._gridsize),
            np.int32(s.shape[2]),
            self.DTYPE_real(self._kwidth / self._gridsize),
            self.dcf.data,
            self.cl_kerneltable,
            np.int32(self._kernelpoints),
            wait_for=(s.events + wait_for + self._tmp_fft_array.events +
                      sg.events))


class PyOpenCLCartNUFFTStreamed(PyOpenCLFFT):
    """ The streamed version of the Cartesian FFT object

    This class performs the FFT operation.

    Attributes:
      fft_shape (tuple of ints):
        3 dimensional tuple. Dim 0 containts all Scans, Coils and Slices.
        Dim 1 and 2 the overgridded image dimensions.
      fft_scale (float32):
        The scaling factor to achieve a good adjointness of the forward and
        backward FFT.
      par_fft (int):
        The number of parallel fft calls. Typically it iterates over the
        Scans.
      fft (gpyfft.fft.FFT):
        The fft object created from gpyfft (A wrapper for clFFT). The object
        is created only once an reused in each iterations, iterationg over
        all scans to keep the memory footprint low.
      mask (PyOpenCL.Array):
        The undersampling mask for the Cartesian grid.
      prg (PyOpenCL.Program):
        The PyOpenCL.Program object containing the necessary kernels to
        execute the linear Operator. This will be determined by the
        factory and set after the object is created.
    """
    def __init__(
            self,
            ctx,
            queue,
            par,
            kwidth=5,
            fft_dim=(
                1,
                2),
            klength=200,
            DTYPE=np.complex64,
            DTYPE_real=np.float32):
        """ Setup the FFT object
        Args:
          ctx (PyOpenCL.Context):
            The context for the PyOpenCL computations.
          queue (PyOpenCL.Queue):
            The computation Queue for the PyOpenCL kernels.
          par (dict): A python dict containing the necessary information to
            setup the object. Needs to contain the number of slices (NSlice),
            number of scans (NScan), image dimensions (dimX, dimY), number of
            coils (NC), sampling points (N) and read outs (NProj)
            a PyOpenCL queue (queue) and the complex coil
            sensitivities (C).
          fft_dim (tuple of ints):
            The dimensions which should be transformed. Defaults
            to 1 and 2 corresponding the the last two of fft_dim.
          DTYPE (Numpy.Type):
            The comlex precission type. Currently complex64 is used.
          DTYPE_real (Numpy.Type):
            The real precission type. Currently float32 is used.
        """
        super().__init__(ctx, queue, DTYPE, DTYPE_real)
        self.fft_shape = (par["NScan"] *
                          par["NC"] *
                          (par["par_slices"] +
                           par["overlap"]), par["dimY"], par["dimX"])
        self.par_fft = int(self.fft_shape[0] / par["NScan"])
        self.fft_scale = DTYPE_real(
            np.sqrt(np.prod(self.fft_shape[fft_dim[0]:])))
        self._tmp_fft_array = (
            clarray.zeros(
                self.queue,
                self.fft_shape,
                dtype=DTYPE))
        self.mask = clarray.to_device(self.queue, par["mask"])
        self.fft = FFT(ctx, queue, self._tmp_fft_array[
            0:self.par_fft, ...],
                       out_array=self._tmp_fft_array[
                           0:self.par_fft, ...],
                       axes=fft_dim)

    def __del__(self):
        del self._tmp_fft_array
        del self.queue
        del self.ctx
        del self.prg
        del self.mask
        del self.fft

    def FFTH(self, sg, s, wait_for=[]):
        """ Perform the inverse (adjoint) FFT operation
        Args:
          sg (PyOpenCL.Array):
            The complex image data.
          s (PyOpenCL.Array):
            The uniformly gridded k-space
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._tmp_fft_array.add_event(
            self.prg.maskingcpy(
                self.queue,
                (self._tmp_fft_array.shape),
                None,
                self._tmp_fft_array.data,
                s.data,
                self.mask.data,
                wait_for=s.events))

        for j in range(int(self.fft_shape[0] / self.par_fft)):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:
                        (j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=False)[0])
        # Scaling
        return (
            self.prg.copy(
                self.queue,
                (sg.size,
                 ),
                None,
                sg.data,
                self._tmp_fft_array.data,
                self.DTYPE_real(
                    self.fft_scale)))

    def FFT(self, s, sg, wait_for=[]):
        """ Perform the forward FFT operation
        Args:
          s (PyOpenCL.Array):
            The uniformly gridded k-space.
          sg (PyOpenCL.Array):
            The complex image data.
          wait_for (list of PyopenCL.Event):
            A List of PyOpenCL events to wait for.
        Returns:
          PyOpenCL.Event: A PyOpenCL event to wait for.
        """
        self._tmp_fft_array.add_event(
            self.prg.copy(
                self.queue,
                (sg.size,
                 ),
                None,
                self._tmp_fft_array.data,
                sg.data,
                self.DTYPE_real(
                    1 /
                    self.fft_scale)))
        for j in range(int(self.fft_shape[0] / self.par_fft)):
            self._tmp_fft_array.add_event(
                self.fft.enqueue_arrays(
                    data=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    result=self._tmp_fft_array[
                        j * self.par_fft:(j + 1) * self.par_fft, ...],
                    forward=True)[0])

        return (
            self.prg.maskingcpy(
                self.queue,
                (self._tmp_fft_array.shape),
                None,
                s.data,
                self._tmp_fft_array.data,
                self.mask.data,
                wait_for=s.events+self._tmp_fft_array.events))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 12 11:26:41 2019

@author: omaier
"""

import pyqmri
try:
    import unittest2 as unittest
except ImportError:
    import unittest
from pyqmri._helper_fun import CLProgram as Program
from pkg_resources import resource_filename
import pyopencl.array as clarray
import numpy as np

DTYPE = np.complex128
DTYPE_real = np.float64


class tmpArgs():
    pass


def setupPar(par):
    par["NScan"] = 10
    par["NC"] = 15
    par["NSlice"] = 20
    par["dimX"] = 128
    par["dimY"] = 128
    par["Nproj"] = 21
    par["N"] = 256
    par["unknowns_TGV"] = 2
    par["unknowns_H1"] = 0
    par["unknowns"] = 2
    par["dz"] = 1
    par["weights"] = [1, 1]


class GradientTest(unittest.TestCase):
    def setUp(self):
        parser = tmpArgs()
        parser.streamed = False
        parser.devices = [0]
        parser.use_GPU = True

        par = {}
        pyqmri.pyqmri._setupOCL(parser, par)
        setupPar(par)
        if DTYPE == np.complex128:
            file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_Kernels_double.c'))
        else:
            file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_Kernels.c'))
        prg = Program(
            par["ctx"][0],
            file.read())
        file.close()

        self.grad = pyqmri.operator.OperatorFiniteGradient(
            par, prg,
            DTYPE=DTYPE,
            DTYPE_real=DTYPE_real)
        self.grad._ratio[0] = 1
        self.grad._ratio[1] = 1
        self.gradin = np.random.randn(par["unknowns"], par["NSlice"],
                                      par["dimY"], par["dimX"]) +\
            1j * np.random.randn(par["unknowns"], par["NSlice"],
                                 par["dimY"], par["dimX"])
        self.divin = np.random.randn(par["unknowns"], par["NSlice"],
                                     par["dimY"], par["dimX"], 4) +\
            1j * np.random.randn(par["unknowns"], par["NSlice"],
                                 par["dimY"], par["dimX"], 4)
        self.gradin = self.gradin.astype(DTYPE)
        self.divin = self.divin.astype(DTYPE)
        self.dz = par["dz"]
        self.queue = par["queue"][0]

    def test_grad_outofplace(self):
        gradx = np.zeros_like(self.gradin)
        grady = np.zeros_like(self.gradin)
        gradz = np.zeros_like(self.gradin)

        gradx[..., :-1] = np.diff(self.gradin, axis=-1)
        grady[..., :-1, :] = np.diff(self.gradin, axis=-2)
        gradz[:, :-1, ...] = np.diff(self.gradin, axis=-3)/self.dz

        grad = np.stack((gradx,
                         grady,
                         gradz), axis=-1)

        inp = clarray.to_device(self.queue, self.gradin)
        outp = self.grad.fwdoop(inp)
        outp = outp.get()

        np.testing.assert_allclose(outp[..., :-1], grad)

    def test_grad_inplace(self):
        gradx = np.zeros_like(self.gradin)
        grady = np.zeros_like(self.gradin)
        gradz = np.zeros_like(self.gradin)

        gradx[..., :-1] = np.diff(self.gradin, axis=-1)
        grady[..., :-1, :] = np.diff(self.gradin, axis=-2)
        gradz[:, :-1, ...] = np.diff(self.gradin, axis=-3)/self.dz

        grad = np.stack((gradx,
                         grady,
                         gradz), axis=-1)

        inp = clarray.to_device(self.queue, self.gradin)
        outp = clarray.to_device(self.queue, self.divin)
        self.grad.fwd(outp, inp)
        outp = outp.get()

        np.testing.assert_allclose(outp[..., :-1], grad)

    def test_adj_outofplace(self):
        inpgrad = clarray.to_device(self.queue, self.gradin)
        inpdiv = clarray.to_device(self.queue, self.divin)

        outgrad = self.grad.fwdoop(inpgrad)
        outdiv = self.grad.adjoop(inpdiv)

        outgrad = outgrad.get()
        outdiv = outdiv.get()

        a = np.vdot(outgrad[..., :-1].flatten(),
                    self.divin[..., :-1].flatten())/self.gradin.size
        b = np.vdot(self.gradin.flatten(), -outdiv.flatten())/self.gradin.size

        print("Adjointness: %.2e +1j %.2e" % ((a - b).real, (a - b).imag))

        self.assertAlmostEqual(a, b, places=14)

    def test_adj_inplace(self):
        inpgrad = clarray.to_device(self.queue, self.gradin)
        inpdiv = clarray.to_device(self.queue, self.divin)

        outgrad = clarray.zeros_like(inpdiv)
        outdiv = clarray.zeros_like(inpgrad)

        self.grad.fwd(outgrad, inpgrad)
        self.grad.adj(outdiv, inpdiv)

        outgrad = outgrad.get()
        outdiv = outdiv.get()

        a = np.vdot(outgrad[..., :-1].flatten(),
                    self.divin[..., :-1].flatten())/self.gradin.size
        b = np.vdot(self.gradin.flatten(), -outdiv.flatten())/self.gradin.size

        print("Adjointness: %.2e +1j %.2e" % ((a - b).real, (a - b).imag))

        self.assertAlmostEqual(a, b, places=14)


class GradientStreamedTest(unittest.TestCase):
    def setUp(self):
        parser = tmpArgs()
        parser.streamed = True
        parser.devices = [0]
        parser.use_GPU = True

        par = {}
        pyqmri.pyqmri._setupOCL(parser, par)
        setupPar(par)
        if DTYPE == np.complex128:
            file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_Kernels_streamed_double.c'))
        else:
            file = open(
                    resource_filename(
                        'pyqmri', 'kernels/OpenCL_Kernels_streamed.c'))

        prg = []
        for j in range(1):
            prg.append(
                Program(
                    par["ctx"][0],
                    file.read()))
        file.close()

        par["par_slices"] = 4

        self.grad = pyqmri.operator.OperatorFiniteGradientStreamed(
            par, prg,
            DTYPE=DTYPE,
            DTYPE_real=DTYPE_real)

        for j in range(len(self.grad._ratio)):
            self.grad._ratio[j][0] = 1
            self.grad._ratio[j][1] = 1

        self.gradin = np.random.randn(par["NSlice"], par["unknowns"],
                                      par["dimY"], par["dimX"]) +\
            1j * np.random.randn(par["NSlice"], par["unknowns"],
                                 par["dimY"], par["dimX"])
        self.divin = np.random.randn(par["NSlice"], par["unknowns"],
                                     par["dimY"], par["dimX"], 4) +\
            1j * np.random.randn(par["NSlice"], par["unknowns"],
                                 par["dimY"], par["dimX"], 4)
        self.gradin = self.gradin.astype(DTYPE)
        self.divin = self.divin.astype(DTYPE)
        self.dz = par["dz"]

    def test_grad_outofplace(self):
        gradx = np.zeros_like(self.gradin)
        grady = np.zeros_like(self.gradin)
        gradz = np.zeros_like(self.gradin)

        gradx[..., :-1] = np.diff(self.gradin, axis=-1)
        grady[..., :-1, :] = np.diff(self.gradin, axis=-2)
        gradz[:-1, ...] = np.diff(self.gradin, axis=0)/self.dz

        grad = np.stack((gradx,
                         grady,
                         gradz), axis=-1)

        outp = self.grad.fwdoop([[self.gradin]])

        np.testing.assert_allclose(outp[..., :-1], grad)

    def test_grad_inplace(self):
        gradx = np.zeros_like(self.gradin)
        grady = np.zeros_like(self.gradin)
        gradz = np.zeros_like(self.gradin)

        gradx[..., :-1] = np.diff(self.gradin, axis=-1)
        grady[..., :-1, :] = np.diff(self.gradin, axis=-2)
        gradz[:-1, ...] = np.diff(self.gradin, axis=0)/self.dz

        grad = np.stack((gradx,
                         grady,
                         gradz), axis=-1)

        outp = np.zeros_like(self.divin)

        self.grad.fwd([outp], [[self.gradin]])

        np.testing.assert_allclose(outp[..., :-1], grad)

    def test_adj_outofplace(self):

        outgrad = self.grad.fwdoop([[self.gradin]])
        outdiv = self.grad.adjoop([[self.divin]])

        a = np.vdot(outgrad[..., :-1].flatten(),
                    self.divin[..., :-1].flatten())/self.gradin.size
        b = np.vdot(self.gradin.flatten(), -outdiv.flatten())/self.gradin.size

        print("Adjointness: %.2e +1j %.2e" % ((a - b).real, (a - b).imag))

        self.assertAlmostEqual(a, b, places=14)

    def test_adj_inplace(self):

        outgrad = np.zeros_like(self.divin)
        outdiv = np.zeros_like(self.gradin)

        self.grad.fwd([outgrad], [[self.gradin]])
        self.grad.adj([outdiv], [[self.divin]])

        a = np.vdot(outgrad[..., :-1].flatten(),
                    self.divin[..., :-1].flatten())/self.gradin.size
        b = np.vdot(self.gradin.flatten(), -outdiv.flatten())/self.gradin.size

        print("Adjointness: %.2e +1j %.2e" % ((a - b).real, (a - b).imag))

        self.assertAlmostEqual(a, b, places=1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from pyqmri.models.template import BaseModel, constraints, DTYPE
plt.ion()


class Model(BaseModel):
    """ Realization of a bi-exponential decay model

      The model assumes M0*(f*exp(-TE*A1)+(1-f)exp(-TE*A2))

    Attributes:
      TE (float): Echo time or parameter in exponential exp(-TR/x)
      guess (numpy.Array): Initial guess
    """
    def __init__(self, par):
        super().__init__(par)

        self.TE = np.ones((self.NScan, 1, 1, 1))
        try:
            for i in range(self.NScan):
                self.TE[i, ...] = par["TE"][i] * np.ones((1, 1, 1))
        except KeyError:
            raise KeyError("No TE found!")

        par["unknowns_TGV"] = 4
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"] + par["unknowns_H1"]

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        self.constraints.append(
            constraints(-100 / self.uk_scale[0],
                        100 / self.uk_scale[0],
                        False))
        self.constraints.append(
                constraints(0 / self.uk_scale[1],
                            1 / self.uk_scale[1],
                            True))
        self.constraints.append(
                constraints(((1 / 150) / self.uk_scale[2]),
                            ((1 / 1e-4) / self.uk_scale[2]),
                            True))
        self.constraints.append(
                constraints(((1 / 1500) / self.uk_scale[3]),
                            ((1 / 150) / self.uk_scale[3]),
                            True))

    def rescale(self, x):
        """ Rescales each unknown by its scaling factor.

          This realization takes into account that A1 and A2 are fitted and
          will return 1/A1 respectively 1/A2 as result.

        Args:
          x (numpy.Array): The array of unknowns to be rescaled
        Returns:
          numpy.Array: The rescaled array of unknowns
        """
        M0 = x[0, ...] * self.uk_scale[0]
        f = x[1, ...] * self.uk_scale[1]
        T21 = 1 / (x[2, ...] * self.uk_scale[2])
        T22 = 1 / (x[3, ...] * self.uk_scale[3])
        return np.array((M0, f, T21, T22))

    def execute_forward_2D(self, x, islice):
        pass

    def execute_gradient_2D(self, x, islice):
        pass

    def execute_forward_3D(self, x):
        M0 = x[0, ...] * self.uk_scale[0]
        f = x[1, ...] * self.uk_scale[1]
        T21 = x[2, ...] * self.uk_scale[2]
        T22 = x[3, ...] * self.uk_scale[3]
        S = M0 * (f * np.exp(-self.TE * (T21)) +
                  (1-f) * np.exp(-self.TE * (T22)))
        S[~np.isfinite(S)] = 1e-20
        S = np.array(S, dtype=DTYPE)
        return S

    def execute_gradient_3D(self, x):
        M0 = x[0, ...]
        f = x[1, ...]
        T21 = x[2, ...]
        T22 = x[3, ...]
        grad_M0 = self.uk_scale[0] * (
            f * self.uk_scale[1] * np.exp(-self.TE * (
                T21 * self.uk_scale[2])) +
            (1 - f * self.uk_scale[1]) *
            np.exp(-self.TE * (T22 * self.uk_scale[3])))

        grad_f = M0 * self.uk_scale[0] * (
            self.uk_scale[1] * np.exp(-self.TE * (T21 * self.uk_scale[2])) -
            self.uk_scale[1] * np.exp(-self.TE * (T22 * self.uk_scale[3])))

        grad_T21 = -self.uk_scale[0] * M0 * f * self.uk_scale[1] * \
            self.TE * \
            self.uk_scale[2] * np.exp(-self.TE * (T21 * self.uk_scale[2]))

        grad_T22 = -self.uk_scale[0] * M0 * (1-f * self.uk_scale[1]) * \
            self.TE * \
            self.uk_scale[3] * np.exp(-self.TE * (T22 * self.uk_scale[4]))
        grad = np.array([grad_M0, grad_f, grad_T21, grad_T22], dtype=DTYPE)
        grad[~np.isfinite(grad)] = 1e-20
        return grad

    def computeInitialGuess(self, *args):
        test_T21 = 1/150 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_T22 = 1/500 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_M0 = np.ones((self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_f = 0.5*np.ones((self.NSlice, self.dimY, self.dimX), dtype=DTYPE)

        x = np.array([test_M0 / self.uk_scale[0],
                      test_f / self.uk_scale[1],
                      test_T21 / self.uk_scale[2],
                      test_T22 / self.uk_scale[3]], dtype=DTYPE)
        self.guess = x

    def plot_unknowns(self, x, dim_2D=False):
        M0 = np.abs(x[0, ...]) * self.uk_scale[0]
        M0_min = M0.min()
        M0_max = M0.max()

        f = np.abs(x[1, ...]) * self.uk_scale[1]

        T21 = 1 / (np.abs(x[2, ...]) * self.uk_scale[2])
        f_min = f.min()
        f_max = f.max()
        T21_min = T21.min()
        T21_max = T21.max()

        T22 = 1 / (np.abs(x[3, ...]) * self.uk_scale[3])
        T22_min = T22.min()
        T22_max = T22.max()

        [z, y, x] = f.shape
        self.ax = []
        if not self.figure:
            plt.ion()
            self.figure = plt.figure(figsize=(12, 6))
            self.figure.subplots_adjust(hspace=0, wspace=0)
            self.gs = gridspec.GridSpec(2,
                                        17,
                                        width_ratios=[x / z,
                                                      1,
                                                      x / (20 * z),
                                                      x / z * 0.25,
                                                      x / (20 * z),
                                                      x / z,
                                                      1,
                                                      x / z,
                                                      1,
                                                      x / (20 * z),
                                                      x / z * 0.25,
                                                      x / (20 * z),
                                                      x / z,
                                                      1,
                                                      x / z,
                                                      1,
                                                      x / (20 * z)],
                                        height_ratios=[x / z,
                                                       1])
            self.figure.tight_layout()
            self.figure.patch.set_facecolor(plt.cm.viridis.colors[0])
            for grid in self.gs:
                self.ax.append(plt.subplot(grid))
                self.ax[-1].axis('off')

            self.M0_plot = self.ax[0].imshow(
                (M0[int(self.NSlice / 2), ...]))
            self.M0_plot_cor = self.ax[17].imshow(
                (M0[:, int(f.shape[1] / 2), ...]))
            self.M0_plot_sag = self.ax[1].imshow(
                np.flip((M0[:, :, int(f.shape[-1] / 2)]).T, 1))
            self.ax[0].set_title('Proton Density in a.u.', color='white')
            self.ax[0].set_anchor('SE')
            self.ax[1].set_anchor('SW')
            self.ax[17].set_anchor('NW')
            cax = plt.subplot(self.gs[:, 2])
            cbar = self.figure.colorbar(self.M0_plot, cax=cax)
            cbar.ax.tick_params(labelsize=12, colors='white')
#           cax.yaxis.set_ticks_position('left')
            for spine in cbar.ax.spines:
                cbar.ax.spines[spine].set_color('white')

            self.f_plot = self.ax[5].imshow(
                (f[int(self.NSlice / 2), ...]))
            self.f_plot_cor = self.ax[22].imshow(
                (f[:, int(f.shape[1] / 2), ...]))
            self.f_plot_sag = self.ax[6].imshow(
                np.flip((f[:, :, int(f.shape[-1] / 2)]).T, 1))
            self.ax[5].set_title('Proton Density in a.u.', color='white')
            self.ax[5].set_anchor('SE')
            self.ax[6].set_anchor('SW')
            self.ax[22].set_anchor('NW')
            cax = plt.subplot(self.gs[:, 4])
            cbar = self.figure.colorbar(self.f_plot, cax=cax)
            cbar.ax.tick_params(labelsize=12, colors='white')
            cax.yaxis.set_ticks_position('left')
            for spine in cbar.ax.spines:
                cbar.ax.spines[spine].set_color('white')

            self.T21_plot = self.ax[7].imshow(
                (T21[int(self.NSlice / 2), ...]))
            self.T21_plot_cor = self.ax[24].imshow(
                (T21[:, int(T21.shape[1] / 2), ...]))
            self.T21_plot_sag = self.ax[8].imshow(
                np.flip((T21[:, :, int(T21.shape[-1] / 2)]).T, 1))
            self.ax[7].set_title('T21 in  ms', color='white')
            self.ax[7].set_anchor('SE')
            self.ax[8].set_anchor('SW')
            self.ax[24].set_anchor('NW')
            cax = plt.subplot(self.gs[:, 9])
            cbar = self.figure.colorbar(self.T21_plot, cax=cax)
            cbar.ax.tick_params(labelsize=12, colors='white')
            for spine in cbar.ax.spines:
                cbar.ax.spines[spine].set_color('white')

            self.T22_plot = self.ax[14].imshow(
                (T22[int(self.NSlice / 2), ...]))
            self.T22_plot_cor = self.ax[31].imshow(
                (T22[:, int(T22.shape[1] / 2), ...]))
            self.T22_plot_sag = self.ax[15].imshow(
                np.flip((T22[:, :, int(T22.shape[-1] / 2)]).T, 1))
            self.ax[14].set_title('T22 in  ms', color='white')
            self.ax[14].set_anchor('SE')
            self.ax[15].set_anchor('SW')
            self.ax[31].set_anchor('NW')
            cax = plt.subplot(self.gs[:, 16])
            cbar = self.figure.colorbar(self.T22_plot, cax=cax)
            cbar.ax.tick_params(labelsize=12, colors='white')
            for spine in cbar.ax.spines:
                cbar.ax.spines[spine].set_color('white')

            plt.draw()
            plt.pause(1e-10)
        else:
            self.M0_plot.set_data((M0[int(self.NSlice / 2), ...]))
            self.M0_plot_cor.set_data((M0[:, int(f.shape[1] / 2), ...]))
            self.M0_plot_sag.set_data(
                np.flip((M0[:, :, int(f.shape[-1] / 2)]).T, 1))
            self.M0_plot.set_clim([M0_min, M0_max])
            self.M0_plot_cor.set_clim([M0_min, M0_max])
            self.M0_plot_sag.set_clim([M0_min, M0_max])

            self.f_plot.set_data((f[int(self.NSlice / 2), ...]))
            self.f_plot_cor.set_data(
                (f[:, int(f.shape[1] / 2), ...]))
            self.f_plot_sag.set_data(
                np.flip((f[:, :, int(f.shape[-1] / 2)]).T, 1))
            self.f_plot.set_clim([f_min, f_max])
            self.f_plot_cor.set_clim([f_min, f_max])
            self.f_plot_sag.set_clim([f_min, f_max])
            self.T21_plot.set_data((T21[int(self.NSlice / 2), ...]))
            self.T21_plot_cor.set_data(
                (T21[:, int(T21.shape[1] / 2), ...]))
            self.T21_plot_sag.set_data(
                np.flip((T21[:, :, int(T21.shape[-1] / 2)]).T, 1))
            self.T21_plot.set_clim([T21_min, T21_max])
            self.T21_plot_sag.set_clim([T21_min, T21_max])
            self.T21_plot_cor.set_clim([T21_min, T21_max])

            self.T22_plot.set_data((T22[int(self.NSlice / 2), ...]))
            self.T22_plot_cor.set_data(
                (T22[:, int(T22.shape[1] / 2), ...]))
            self.T22_plot_sag.set_data(
                np.flip((T22[:, :, int(T22.shape[-1] / 2)]).T, 1))
            self.T22_plot.set_clim([T22_min, T22_max])
            self.T22_plot_sag.set_clim([T22_min, T22_max])
            self.T22_plot_cor.set_clim([T22_min, T22_max])
            plt.draw()
            plt.pause(1e-10)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import division

import numpy as np
import time
import sys
from pkg_resources import resource_filename
import pyopencl.array as clarray
import h5py
import pyqmri.operator as operator
import pyqmri.streaming as streaming
from pyqmri._helper_fun import CLProgram as Program
from pyqmri._helper_fun import _utils as utils
DTYPE = np.complex64
DTYPE_real = np.float32


class ModelReco:
    def __init__(self, par, trafo=1, imagespace=False, reg_type='TGV',
                 config='', model=None):

        par["overlap"] = 1
        self.overlap = par["overlap"]
        self.par_slices = par["par_slices"]
        self.par = par
        self.C = np.require(
            np.transpose(par["C"], [1, 0, 2, 3]), requirements='C',
            dtype=DTYPE)
        self.unknowns_TGV = par["unknowns_TGV"]
        self.unknowns_H1 = par["unknowns_H1"]
        self.unknowns = par["unknowns"]
        self.NSlice = par["NSlice"]
        self.NScan = par["NScan"]
        self.dimX = par["dimX"]
        self.dimY = par["dimY"]
        self.scale = 1

        self.N = par["N"]
        self.Nproj = par["Nproj"]
        self.dz = 1
        self.fval_old = 0
        self.fval = 0
        self.fval_init = 0
        self.SNR_est = par["SNR_est"]
        self.ctx = par["ctx"]
        self.queue = par["queue"]
        self.gn_res = []
        self.irgn_par = utils.read_config(config, reg_type)
        self.model = model
        self.reg_type = reg_type
        self.num_dev = len(par["num_dev"])
        if self.NSlice/(self.num_dev*self.par_slices) < 2:
            raise ValueError(
                "Number of Slices devided by parallel "
                "computed slices and devices needs to be larger two.")
        if self.NSlice % self.par_slices:
            raise ValueError(
                "Number of Slices devided by parallel "
                "computed slices needs to be an integer.")
        self.prg = []
        for j in range(self.num_dev):
            self.prg.append(
                Program(
                    self.ctx[j],
                    open(
                        resource_filename(
                            'pyqmri',
                            'kernels/OpenCL_Kernels_streamed.c')).read()))

        self.tmp_img = []

        self.unknown_shape = (self.NSlice, self.unknowns, self.dimY, self.dimX)
        self.grad_shape = self.unknown_shape + (4,)
        self._imagespace = imagespace
        if imagespace:
            self.data_shape = (self.NSlice, self.NScan,
                               self.dimY, self.dimX)
            self.C = []
            self.NC = 1
            self.N = self.dimX
            self.Nproj = self.dimY
            self.dat_trans_axes = [1, 0, 2, 3]
            self.op = operator.OperatorImagespaceStreamed(par, self.prg)
        else:
            self.NC = par["NC"]
            self.dat_trans_axes = [2, 0, 1, 3, 4]

            self.data_shape = (self.NSlice, self.NScan,
                               self.NC, self.Nproj, self.N)
            self.data_shape_T = self.data_shape
            self.op = operator.OperatorKspaceStreamed(par,
                                                      self.prg,
                                                      trafo=trafo)
            self._expdim_dat = 2
            self._expdim_C = 1

        self.grad_op, self.symgrad_op, self.v = self._setupLinearOps()

        self.step_val = None
        self.pdop = None
        self.model_partial_der = None
        self.grad_buf = None
        self.delta = None
        self.delta_max = None
        self.omega = None
        self.gamma = None
        self.data = None  # Needs to be set outside

    def update_Kyk2(self, outp, inp, par=None, idx=0, idxq=0,
                    bound_cond=0, wait_for=[]):
        return self.prg[idx].update_Kyk2(
            self.queue[4*idx+idxq],
            (self.overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, inp[1].data,
            np.int32(self.unknowns), np.int32(bound_cond), np.float32(self.dz),
            wait_for=outp.events + inp[0].events + inp[1].events+wait_for)

    def update_primal(self, outp, inp, par=None, idx=0, idxq=0,
                      bound_cond=0, wait_for=[]):
        return self.prg[idx].update_primal(
            self.queue[4*idx+idxq],
            (self.overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, inp[1].data, inp[2].data,
            np.float32(par[0]),
            np.float32(par[0]/par[1]), np.float32(1/(1+par[0]/par[1])),
            self.min_const[idx].data, self.max_const[idx].data,
            self.real_const[idx].data, np.int32(self.unknowns),
            wait_for=(outp.events +
                      inp[0].events+inp[1].events +
                      inp[2].events+wait_for))

    def update_z1(self, outp, inp, par=None, idx=0, idxq=0,
                  bound_cond=0, wait_for=[]):
        return self.prg[idx].update_z1(
            self.queue[4*idx+idxq],
            (self.overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, inp[1].data,
            inp[2].data, inp[3].data, inp[4].data,
            np.float32(par[0]), np.float32(par[1]),
            np.float32(1/par[2]), np.int32(self.unknowns_TGV),
            np.int32(self.unknowns_H1), np.float32(1 / (1 + par[0] / par[3])),
            wait_for=(outp.events+inp[0].events+inp[1].events +
                      inp[2].events+inp[3].events+inp[4].events+wait_for))

    def update_z1_tv(self, outp, inp, par=None, idx=0, idxq=0,
                     bound_cond=0, wait_for=[]):
        return self.prg[idx].update_z1_tv(
            self.queue[4*idx+idxq],
            (self.overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, inp[0].data, inp[0].data,
            np.float32(par[0]),
            np.float32(par[1]),
            np.float32(1/par[2]), np.int32(self.unknowns_TGV),
            np.int32(self.unknowns_H1), np.float32(1 / (1 + par[0] / par[3])),
            wait_for=(outp.events+inp[0].events +
                      inp[1].events+inp[2].events+wait_for))

    def update_z2(self, outp, inp, par=None, idx=0, idxq=0,
                  bound_cond=0, wait_for=[]):
        return self.prg[idx].update_z2(
            self.queue[4*idx+idxq],
            (self.overlap+self.par_slices, self.dimY, self.dimX), None,
            outp.data, inp[0].data, inp[1].data, inp[2].data,
            np.float32(par[0]),
            np.float32(par[1]),
            np.float32(1/par[2]), np.int32(self.unknowns),
            wait_for=(outp.events+inp[0].events +
                      inp[1].events+inp[2].events+wait_for))

    def update_r(self, outp, inp, par=None, idx=0, idxq=0,
                 bound_cond=0, wait_for=[]):
        return self.prg[idx].update_r(
            self.queue[4*idx+idxq], (outp.size,), None,
            outp.data, inp[0].data,
            inp[1].data, inp[2].data, inp[3].data,
            np.float32(par[0]), np.float32(par[1]),
            np.float32(1/(1+par[0]/self.irgn_par["lambd"])),
            wait_for=(outp.events+inp[0].events +
                      inp[1].events+inp[2].events+wait_for))

    def update_v(self, outp, inp, par=None, idx=0, idxq=0,
                 bound_cond=0, wait_for=[]):
        return self.prg[idx].update_v(
            self.queue[4*idx+idxq], (outp[..., 0].size,), None,
            outp.data, inp[0].data, inp[1].data, np.float32(par[0]),
            wait_for=outp.events+inp[0].events+inp[1].events+wait_for)

    def eval_const(self):
        num_const = (len(self.model.constraints))
        min_const = np.zeros((num_const), dtype=np.float32)
        max_const = np.zeros((num_const), dtype=np.float32)
        real_const = np.zeros((num_const), dtype=np.int32)
        for j in range(num_const):
            min_const[j] = np.float32(self.model.constraints[j].min)
            max_const[j] = np.float32(self.model.constraints[j].max)
            real_const[j] = np.int32(self.model.constraints[j].real)

        self.min_const = []
        self.max_const = []
        self.real_const = []
        for j in range(self.num_dev):
            self.min_const.append(
                clarray.to_device(self.queue[4*j], min_const))
            self.max_const.append(
                clarray.to_device(self.queue[4*j], max_const))
            self.real_const.append(
                clarray.to_device(self.queue[4*j], real_const))

    def _setupLinearOps(self):
        grad_op = operator.OperatorFiniteGradientStreamed(self.par, self.prg)
        symgrad_op = None
        v = None
        if self.reg_type == 'TGV':
            symgrad_op = operator.OperatorFiniteSymGradientStreamed(
                self.par, self.prg)
            v = np.zeros(
                ([self.par["unknowns"], self.par["NSlice"],
                  self.par["dimY"], self.par["dimX"], 4]),
                dtype=DTYPE)
        return grad_op, symgrad_op, v

    def execute(self, reco_2D=0):
        if reco_2D:
            NotImplementedError("2D currently not implemented, "
                                "3D can be used with a single slice.")
        else:
            self.irgn_par["lambd"] *= 1e2/np.sqrt(self.par["SNR_est"])
            self.delta = self.irgn_par["delta"]
            self.delta_max = self.irgn_par["delta_max"]
            self.gamma = self.irgn_par["gamma"]
            self.omega = self.irgn_par["omega"]
            self._setup_reg_tmp_arrays()
            self.execute_3D()

###############################################################################
# Start a 3D Reconstruction, set TV to True to perform TV instead of TGV#######
# Precompute Model and Gradient values for xk #################################
# Call inner optimization #####################################################
# input: bool to switch between TV (1) and TGV (0) regularization #############
# output: optimal value of x ##################################################
###############################################################################
    def execute_3D(self):

        iters = self.irgn_par["start_iters"]
        result = np.copy(self.model.guess)
        self.data = np.require(
            np.transpose(self.data, self.dat_trans_axes), requirements='C')
        self.grad_x = np.nan_to_num(self.model.execute_gradient(result))

        self._balanceModelGradients(result, 0)
        self.grad_op.updateRatio(result)
        self._updateIRGNRegPar(result, 0)

        for ign in range(self.irgn_par["max_gn_it"]):
            start = time.time()

            if ign > 0:
                self.grad_x = np.nan_to_num(
                    self.model.execute_gradient(result))
                self._balanceModelGradients(result, ign)
                self.grad_op.updateRatio(result)

            self.step_val = np.nan_to_num(
                self.model.execute_forward(result))

            self.step_val = np.require(
                np.transpose(
                    self.step_val, [1, 0, 2, 3]), requirements='C')

            self.grad_x = np.require(
                np.transpose(self.grad_x, [2, 0, 1, 3, 4]), requirements='C')

            self._updateIRGNRegPar(result, ign)

            result = self.irgn_solve_3D(result, iters, ign)

            iters = np.fmin(iters * 2, self.irgn_par["max_iters"])

            end = time.time() - start

            self.gn_res.append(self.fval)
            print("-" * 75)
            print("GN-Iter: %d  Elapsed time: %f seconds" % (ign, end))
            print("-" * 75)
            self.fval_old = self.fval
            self.savetofile(ign, self.model.rescale(result))

        self.calc_residual(
            np.require(
                np.transpose(result, [1, 0, 2, 3]),
                requirements='C'),
            ign+1)

    def _updateIRGNRegPar(self, result, ign):
        self.irgn_par["delta_max"] = (self.delta_max /
                                      1e3 *
                                      np.linalg.norm(result))
        self.irgn_par["delta"] = np.minimum(
            self.delta /
            (1e3) *
            np.linalg.norm(result)*self.irgn_par["delta_inc"]**ign,
            self.irgn_par["delta_max"])
        self.irgn_par["gamma"] = np.maximum(
            self.gamma * self.irgn_par["gamma_dec"]**ign,
            self.irgn_par["gamma_min"])
        self.irgn_par["omega"] = np.maximum(
            self.omega * self.irgn_par["omega_dec"]**ign,
            self.irgn_par["omega_min"])

    def _balanceModelGradients(self, result, ind):
        scale = np.reshape(
            self.grad_x,
            (self.par["unknowns"],
             self.par["NScan"] * self.par["NSlice"] *
             self.par["dimY"] * self.par["dimX"]))
        scale = np.linalg.norm(scale, axis=-1)
        scale = 1e3 / scale
        if not np.mod(ind, 1):
            for uk in range(self.par["unknowns"]):
                self.model.constraints[uk].update(scale[uk])
                result[uk, ...] *= self.model.uk_scale[uk]
                self.grad_x[uk] /= self.model.uk_scale[uk]
                self.model.uk_scale[uk] *= scale[uk]
                result[uk, ...] /= self.model.uk_scale[uk]
                self.grad_x[uk] *= self.model.uk_scale[uk]

    def _setup_reg_tmp_arrays(self):
        if self.reg_type == 'TV':
            self.tau = np.float32(1/np.sqrt(8))
            self.beta_line = 400
            self.theta_line = np.float32(1.0)
        elif self.reg_type == 'TGV':
            L = np.float32(0.5*(18.0 + np.sqrt(33)))
            self.tau = np.float32(1/np.sqrt(L))
            self.beta_line = 400
            self.theta_line = np.float32(1.0)
            self.v = np.zeros(
                ([self.NSlice, self.unknowns, self.dimY, self.dimX, 4]),
                dtype=DTYPE)
            self.z2 = np.zeros(
                ([self.NSlice, self.unknowns, self.dimY, self.dimX, 8]),
                dtype=DTYPE)
        else:
            raise NotImplementedError("Not implemented")
        self._setupstreamingops()

        self.r = np.zeros_like(self.data, dtype=DTYPE)
        self.r = np.require(np.transpose(self.r, self.dat_trans_axes),
                            requirements='C')
        self.z1 = np.zeros(
            ([self.NSlice, self.unknowns, self.dimY, self.dimX, 4]),
            dtype=DTYPE)

###############################################################################
# New .hdf5 save files ########################################################
###############################################################################
    def savetofile(self, myit, result):
        f = h5py.File(self.par["outdir"]+"output_" + self.par["fname"], "a")
        if self.reg_type == 'TGV':
            f.create_dataset("tgv_result_iter_"+str(myit), result.shape,
                             dtype=DTYPE, data=result)
            f.attrs['res_tgv_iter_'+str(myit)] = self.fval
        else:
            f.create_dataset("tv_result_"+str(myit), result.shape,
                             dtype=DTYPE, data=result)
            f.attrs['res_tv_iter_'+str(myit)] = self.fval
        f.close()

###############################################################################
# Precompute constant terms of the GN linearization step ######################
# input: linearization point x ################################################
# numeber of innner iterations iters ##########################################
# Data ########################################################################
# bool to switch between TV (1) and TGV (0) regularization ####################
# output: optimal value of x for the inner GN step ############################
###############################################################################
###############################################################################
    def irgn_solve_3D(self, x, iters, GN_it):
        x = np.require(np.transpose(x, [1, 0, 2, 3]), requirements='C')
        DGk = np.zeros(self.data_shape, dtype=DTYPE)

        self.op.fwd(
            [DGk],
            [[x, self.C, self.grad_x]])

        b = self.calc_residual(x, GN_it)

        res = self.data - b + DGk

        if self.reg_type == 'TV':
            x = self.tv_solve_3D(x, res, iters)
        elif self.reg_type == 'TGV':
            x = self.tgv_solve_3D(x, res, iters)
        x = np.require(np.transpose(x, [1, 0, 2, 3]), requirements='C')
        return x

    def calc_residual(self, x, GN_it):
        grad = np.zeros(self.z1.shape, dtype=DTYPE)
        self.grad_op.fwd([grad], [[x]])

        if self._imagespace is False:
            b = np.zeros(self.data_shape_T, dtype=DTYPE)
            self.op.FTstr.eval(
                [b],
                [[np.expand_dims(self.step_val, self._expdim_dat) *
                  np.expand_dims(self.C, self._expdim_C)]])
        else:
            b = self.step_val

        datacost = self.irgn_par["lambd"] / 2 *\
            np.linalg.norm(self.data - b)**2
        L2Cost = np.linalg.norm(x)/(2.0*self.irgn_par["delta"])

        if self.reg_type == 'TV':
            regcost = self.irgn_par["gamma"]*np.sum(np.abs(
                    grad[:, :self.unknowns_TGV]))

        elif self.reg_type == 'TGV':
            sym_grad = np.zeros_like(self.z2)
            self.symgrad_op.fwd([sym_grad], [[self.v]])
            regcost = self.irgn_par["gamma"] * np.sum(
                  np.abs(grad[:, :self.par["unknowns_TGV"]] -
                         self.v)) + self.irgn_par["gamma"] * 2 * np.sum(
                             np.abs(sym_grad))
            del sym_grad

        self.fval = (datacost +
                     regcost +
                     self.irgn_par["omega"] / 2 *
                     np.linalg.norm(grad[:, self.par["unknowns_TGV"]:])**2)
        del grad
        if GN_it == 0:
            self.fval_init = self.fval
        print("-" * 75)
        print("Costs of Data: %f" % (datacost))
        print("Costs of T(G)V: %f" % (regcost))
        print("Costs of L2 Term: %f" % (L2Cost))
        print("-" * 75)
        print("Function value at GN-Step %i: %f" %
              (GN_it, 1e3*self.fval / self.fval_init))
        print("-" * 75)
        return b

    def tgv_solve_3D(self, x, res, iters):
        alpha = self.irgn_par["gamma"]
        beta = self.irgn_par["gamma"] * 2

        tau = self.tau
        tau_new = np.float32(0)

        xk = x.copy()
        x_new = np.zeros_like(x)

        r = np.zeros_like(self.r)
        r_new = np.zeros_like(r)
        z1 = np.zeros_like(self.z1)
        z1_new = np.zeros_like(z1)
        z2 = np.zeros_like(self.z2)
        z2_new = np.zeros_like(z2)
        v = np.zeros_like(self.v)
        v_new = np.zeros_like(v)
        res = (res).astype(DTYPE)

        delta = self.irgn_par["delta"]
        omega = self.irgn_par["omega"]
        mu = 1/delta

        theta_line = self.theta_line
        beta_line = self.beta_line
        beta_new = np.float32(0)
        mu_line = np.float32(0.5)
        delta_line = np.float32(1)
        ynorm1 = np.float32(0.0)
        lhs1 = np.float32(0.0)
        ynorm2 = np.float32(0.0)
        lhs2 = np.float32(0.0)
        primal = np.float32(0.0)
        primal_new = np.float32(0)
        dual = np.float32(0.0)
        gap_init = np.float32(0.0)
        gap_old = np.float32(0.0)
        gap = np.float32(0.0)
        self.eval_const()

        Kyk1 = np.zeros_like(x)
        Kyk1_new = np.zeros_like(x)
        Kyk2 = np.zeros_like(z1)
        Kyk2_new = np.zeros_like(z1)
        gradx = np.zeros_like(z1)
        gradx_xold = np.zeros_like(z1)
        symgrad_v = np.zeros_like(z2)
        symgrad_v_vold = np.zeros_like(z2)
        Axold = np.zeros_like(res)
        Ax = np.zeros_like(res)

        # Warmup
        self.stream_initial_1.eval(
            [Axold, Kyk1, symgrad_v_vold],
            [[x, self.C, self.grad_x], [r, z1, self.C, self.grad_x, []], [v]],
            [self.grad_op._ratio])
        self.stream_initial_2.eval(
            [gradx_xold, Kyk2],
            [[x], [z2, z1, []]])
        # Start Iterations
        for myit in range(iters):
            self.update_primal_1.eval(
                [x_new, gradx, Ax],
                [[x, Kyk1, xk], [], [[], self.C, self.grad_x]],
                [tau, delta])
            self.update_primal_2.eval(
                [v_new, symgrad_v],
                [[v, Kyk2], []],
                [tau])

            beta_new = beta_line*(1+mu*tau)
            tau_new = tau*np.sqrt(beta_line/beta_new*(1+theta_line))
            beta_line = beta_new

            while True:
                theta_line = tau_new/tau
                (lhs1, ynorm1) = self.update_dual_1.evalwithnorm(
                    [z1_new, r_new, Kyk1_new],
                    [[z1, gradx, gradx_xold, v_new, v],
                     [r, Ax, Axold, res],
                     [[], [], self.C, self.grad_x, Kyk1]],
                    [beta_line*tau_new, theta_line,
                     alpha, omega, self.grad_op._ratio])
                (lhs2, ynorm2) = self.update_dual_2.evalwithnorm(
                    [z2_new, Kyk2_new],
                    [[z2, symgrad_v, symgrad_v_vold], [[], z1_new, Kyk2]],
                    [beta_line*tau_new, theta_line, beta])

                if np.sqrt(beta_line)*tau_new*(abs(lhs1+lhs2)**(1/2)) <= \
                   (abs(ynorm1+ynorm2)**(1/2))*delta_line:
                    break
                else:
                    tau_new = tau_new*mu_line

            (Kyk1, Kyk1_new, Kyk2, Kyk2_new, Axold, Ax, z1, z1_new,
             z2, z2_new, r, r_new, gradx_xold, gradx, symgrad_v_vold,
             symgrad_v, tau) = (
             Kyk1_new, Kyk1, Kyk2_new, Kyk2, Ax, Axold, z1_new, z1,
             z2_new, z2, r_new, r, gradx, gradx_xold, symgrad_v,
             symgrad_v_vold, tau_new)

            if not np.mod(myit, 10):
                if self.irgn_par["display_iterations"]:
                    self.model.plot_unknowns(
                        np.transpose(x_new, [1, 0, 2, 3]))
                if self.unknowns_H1 > 0:
                    primal_new = (
                        self.irgn_par["lambd"]/2 *
                        np.vdot(Axold-res, Axold-res) +
                        alpha*np.sum(abs((gradx[:, :self.unknowns_TGV]-v))) +
                        beta*np.sum(abs(symgrad_v)) +
                        1/(2*delta)*np.vdot(x_new-xk, x_new-xk) +
                        self.irgn_par["omega"] / 2 *
                        np.vdot(gradx[:, :self.unknowns_TGV],
                                gradx[:, :self.unknowns_TGV])).real

                    dual = (
                        - delta/2*np.vdot(-Kyk1.flatten(), -Kyk1.flatten())
                        - np.vdot(xk.flatten(), -Kyk1.flatten())
                        + np.sum(Kyk2)
                        - 1/(2*self.irgn_par["lambd"])
                        * np.vdot(r.flatten(), r.flatten())
                        - np.vdot(res.flatten(), r.flatten())
                        - 1 / (2 * self.irgn_par["omega"])
                        * np.vdot(z1[:, :self.unknowns_TGV],
                                  z1[:, :self.unknowns_TGV])).real
                else:
                    primal_new = (
                        self.irgn_par["lambd"]/2 *
                        np.vdot(Axold-res, Axold-res) +
                        alpha*np.sum(abs((gradx[:, :self.unknowns_TGV]-v))) +
                        beta*np.sum(abs(symgrad_v)) +
                        1/(2*delta)*np.vdot(x_new-xk, x_new-xk)).real

                    dual = (
                        - delta/2*np.vdot(-Kyk1.flatten(), -Kyk1.flatten())
                        - np.vdot(xk.flatten(), -Kyk1.flatten())
                        + np.sum(Kyk2)
                        - 1/(2*self.irgn_par["lambd"])
                        * np.vdot(r.flatten(), r.flatten())
                        - np.vdot(res.flatten(), r.flatten())).real

                gap = np.abs(primal_new - dual)
                if myit == 0:
                    gap_init = gap
                if np.abs((primal-primal_new) / self.fval_init) <\
                   self.irgn_par["tol"]:
                    print("Terminated at iteration %d because the energy "
                          "decrease in the primal problem was less than %.3e" %
                          (myit, np.abs(primal-primal_new) / self.fval_init))
                    self.v = v_new
                    self.r = r
                    self.z1 = z1
                    self.z2 = z2
                    return x_new
                if (gap > gap_old*self.irgn_par["stag"]) and myit > 1:
                    self.v = v_new
                    self.r = r
                    self.z1 = z1
                    self.z2 = z2
                    print("Terminated at iteration %d "
                          "because the method stagnated" % (myit))
                    return x_new
                if np.abs((gap-gap_old)/gap_init) < self.irgn_par["tol"]:
                    self.v = v_new
                    self.r = r
                    self.z1 = z1
                    self.z2 = z2
                    print("Terminated at iteration %d because the relative "
                          "energy decrease of the PD gap was less than %.3e" %
                          (myit, np.abs((gap-gap_old) / gap_init)))
                    return x_new
                primal = primal_new
                gap_old = gap
                sys.stdout.write(
                    "Iteration: %04d ---- Primal: "
                    "%2.2e, Dual: %2.2e, Gap: %2.2e \r"
                    % (myit, 1000*primal/self.fval_init,
                       1000*dual/self.fval_init,
                       1000*gap/self.fval_init))
                sys.stdout.flush()
            (x, x_new) = (x_new, x)
            (v, v_new) = (v_new, v)

        self.v = v
        self.r = r
        self.z1 = z1
        self.z2 = z2
        return x

    def tv_solve_3D(self, x, res, iters):
        alpha = self.irgn_par["gamma"]
        tau = self.tau
        tau_new = np.float32(0)

        xk = x.copy()
        x_new = np.zeros_like(x)

        r = np.zeros_like(self.r)
        r_new = np.zeros_like(r)
        z1 = np.zeros_like(self.z1)
        z1_new = np.zeros_like(z1)
        res = (res).astype(DTYPE)

        delta = self.irgn_par["delta"]
        omega = self.irgn_par["omega"]
        mu = 1/delta
        theta_line = np.float32(1.0)
        beta_line = np.float32(400)
        beta_new = np.float32(0)
        mu_line = np.float32(0.5)
        delta_line = np.float32(1)
        ynorm1 = np.float32(0.0)
        lhs1 = np.float32(0.0)
        ynorm2 = np.float32(0.0)
        lhs2 = np.float32(0.0)
        primal = np.float32(0.0)
        primal_new = np.float32(0)
        dual = np.float32(0.0)
        gap_init = np.float32(0.0)
        gap_old = np.float32(0.0)

        self.eval_const()

        Kyk1 = np.zeros_like(x)
        Kyk1_new = np.zeros_like(x)
        gradx = np.zeros_like(z1)
        gradx_xold = np.zeros_like(z1)
        Axold = np.zeros_like(res)
        Ax = np.zeros_like(res)

        # Warmup
        self.stream_initial_1.eval(
            [Axold, Kyk1],
            [[x, self.C, self.grad_x], [r, z1, self.C, self.grad_x, []]])
        self.stream_initial_2.eval(
            [gradx_xold],
            [[x]])

        for myit in range(iters):
            self.update_primal_1.eval(
                [x_new, gradx, Ax],
                [[x, Kyk1, xk], [], [[], self.C, self.grad_x]],
                [tau, delta])

            beta_new = beta_line*(1+mu*tau)
            tau_new = tau*np.sqrt(beta_line/beta_new*(1+theta_line))
            beta_line = beta_new

            while True:
                theta_line = tau_new/tau

                (lhs1, ynorm1) = self.update_dual_1.evalwithnorm(
                    [z1_new, r_new, Kyk1_new],
                    [[z1, gradx, gradx_xold],
                     [r, Ax, Axold, res],
                     [[], [], self.C, self.grad_x, Kyk1]],
                    [beta_line*tau_new, theta_line,
                     alpha, omega])

                if np.sqrt(beta_line)*tau_new*(abs(lhs1+lhs2)**(1/2)) <= \
                   (abs(ynorm1+ynorm2)**(1/2))*delta_line:
                    break
                else:
                    tau_new = tau_new*mu_line

            (Kyk1, Kyk1_new, Axold, Ax, z1, z1_new, r, r_new, gradx_xold,
             gradx, tau) = (
             Kyk1_new, Kyk1, Ax, Axold, z1_new, z1, r_new, r, gradx,
             gradx_xold, tau_new)

            if not np.mod(myit, 10):
                if self.irgn_par["display_iterations"]:
                    self.model.plot_unknowns(np.transpose(x_new, [1, 0, 2, 3]))
                if self.unknowns_H1 > 0:
                    primal_new = (
                        self.irgn_par["lambd"]/2 *
                        np.vdot(Axold-res, Axold-res) +
                        alpha*np.sum(abs((gradx[:, :self.unknowns_TGV]))) +
                        1/(2*delta)*np.vdot(x_new-xk, x_new-xk) +
                        self.irgn_par["omega"] / 2 *
                        np.vdot(gradx[:, :self.unknowns_TGV],
                                gradx[:, :self.unknowns_TGV])).real

                    dual = (
                        -delta/2*np.vdot(Kyk1, Kyk1) - np.vdot(xk, (-1)*Kyk1)
                        - 1/(2*self.irgn_par["lambd"])*np.vdot(r, r)
                        - np.vdot(res, r)
                        - 1 / (2 * self.irgn_par["omega"])
                        * np.vdot(z1[:, :self.unknowns_TGV],
                                  z1[:, :self.unknowns_TGV])).real
                else:
                    primal_new = (
                        self.irgn_par["lambd"]/2 *
                        np.vdot(Axold-res, Axold-res) +
                        alpha*np.sum(abs((gradx[:, :self.unknowns_TGV]))) +
                        1/(2*delta)*np.vdot(x_new-xk, x_new-xk)).real

                    dual = (
                        -delta/2*np.vdot(Kyk1, Kyk1) - np.vdot(xk, (-1)*Kyk1)
                        - 1/(2*self.irgn_par["lambd"])*np.vdot(r, r)
                        - np.vdot(res, r)).real

                gap = np.abs(primal_new - dual)
                if myit == 0:
                    gap_init = gap
                if np.abs(primal-primal_new) / self.fval_init < \
                   self.irgn_par["tol"]:
                    print("Terminated at iteration %d because the energy "
                          "decrease in the primal problem was less than %.3e" %
                          (myit, np.abs(primal-primal_new) / self.fval_init))
                    self.r = r
                    self.z1 = z1
                    return x_new
                if (gap > gap_old*self.irgn_par["stag"]) and myit > 1:
                    self.r = r
                    self.z1 = z1
                    print("Terminated at iteration %d because "
                          "the method stagnated" % (myit))
                    return x_new
                if np.abs((gap-gap_old)/gap_init) < self.irgn_par["tol"]:
                    self.r = r
                    self.z1 = z1
                    print("Terminated at iteration %d because the relative "
                          "energy decrease of the PD gap was less than %.3e" %
                          (myit, np.abs((gap-gap_old) / gap_init)))
                    return x_new
                primal = primal_new
                gap_old = gap
                sys.stdout.write(
                    "Iteration: %04d ---- "
                    "Primal: %2.2e, Dual: %2.2e, Gap: %2.2e \r"
                    % (myit, 1000 * primal / self.fval_init,
                       1000 * dual / self.fval_init,
                       1000 * gap / self.fval_init))
                sys.stdout.flush()
            (x, x_new) = (x_new, x)

        self.r = r
        self.z1 = z1
        return x

    def _setupstreamingops(self):
        if self.reg_type == 'TGV':
            symgrad_shape = self.unknown_shape + (8,)

        self.stream_initial_1 = self._defineoperator(
            [],
            [],
            [[]],
            reverse_dir=True)
        self.stream_initial_1 += self.op.fwdstr
        self.stream_initial_1 += self.op.adjstrKyk1
        if self.reg_type == 'TGV':
            self.stream_initial_1 += self.symgrad_op._stream_symgrad

        if self.reg_type == 'TGV':
            self.stream_Kyk2 = self._defineoperator(
                [self.update_Kyk2],
                [self.grad_shape],
                [[symgrad_shape,
                  self.grad_shape,
                  self.grad_shape]])

            self.stream_initial_2 = self._defineoperator(
                [],
                [],
                [[]])

            self.stream_initial_2 += self.grad_op._stream_grad
            self.stream_initial_2 += self.stream_Kyk2
        self.stream_primal = self._defineoperator(
            [self.update_primal],
            [self.unknown_shape],
            [[self.unknown_shape,
              self.unknown_shape,
              self.unknown_shape]])

        self.update_primal_1 = self._defineoperator(
            [],
            [],
            [[]])

        self.update_primal_1 += self.stream_primal
        self.update_primal_1 += self.grad_op._stream_grad
        self.update_primal_1 += self.op.fwdstr

        self.update_primal_1.connectouttoin(0, (1, 0))
        self.update_primal_1.connectouttoin(0, (2, 0))

        if self.reg_type == 'TGV':
            self.stream_update_v = self._defineoperator(
                [self.update_v],
                [self.grad_shape],
                [[self.grad_shape,
                  self.grad_shape]])

            self.update_primal_2 = self._defineoperator(
                [],
                [],
                [[]],
                reverse_dir=True)

            self.update_primal_2 += self.stream_update_v
            self.update_primal_2 += self.symgrad_op._stream_symgrad
            self.update_primal_2.connectouttoin(0, (1, 0))

        if self.reg_type == 'TV':
            self.stream_update_z1 = self._defineoperator(
                [self.update_z1],
                [self.grad_shape],
                [[self.grad_shape,
                  self.grad_shape,
                  self.grad_shape]])
        else:
            self.stream_update_z1 = self._defineoperator(
                [self.update_z1],
                [self.grad_shape],
                [[self.grad_shape,
                  self. grad_shape,
                  self. grad_shape,
                  self. grad_shape,
                  self. grad_shape]])

        self.stream_update_r = self._defineoperator(
            [self.update_r],
            [self.data_shape],
            [[self.data_shape,
              self.data_shape,
              self.data_shape,
              self.data_shape]])

        self.update_dual_1 = self._defineoperator(
            [],
            [],
            [[]],
            reverse_dir=True,
            posofnorm=[False, False, True])

        self.update_dual_1 += self.stream_update_z1
        self.update_dual_1 += self.stream_update_r
        self.update_dual_1 += self.op.adjstrKyk1
        self.update_dual_1.connectouttoin(0, (2, 1))
        self.update_dual_1.connectouttoin(1, (2, 0))

        del self.stream_update_z1, self.stream_update_r, \
            self.stream_update_v, self.stream_primal

        if self.reg_type == 'TGV':
            self.stream_update_z2 = self._defineoperator(
                [self.update_z2],
                [symgrad_shape],
                [[symgrad_shape,
                  symgrad_shape,
                  symgrad_shape]])

            self.update_dual_2 = self._defineoperator(
                [],
                [],
                [[]],
                posofnorm=[False, True])

            self.update_dual_2 += self.stream_update_z2
            self.update_dual_2 += self.stream_Kyk2
            self.update_dual_2.connectouttoin(0, (1, 0))
            del self.stream_Kyk2, self.stream_update_z2

    def _defineoperator(self,
                        functions,
                        outp,
                        inp,
                        reverse_dir=False,
                        posofnorm=[],
                        slices=None):
        if slices is None:
            slices = self.NSlice
        return streaming.Stream(
            functions,
            outp,
            inp,
            self.par_slices,
            self.overlap,
            slices,
            self.queue,
            self.num_dev,
            reverse_dir,
            posofnorm)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from pyqmri.models.template import BaseModel, constraints, DTYPE
plt.ion()



class Model(BaseModel):
    def __init__(self, par):
        super().__init__(par)
        self.constraints = []
        self.TR = par["TR"]
        self.fa = par["flip_angle(s)"]
        try:
            self.fa_corr = par["fa_corr"]
        except KeyError:
            self.fa_corr = 1
            print("No flipangle correction found!")

        phi_corr = np.zeros(
          (self.NScan, self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        for i in range(np.size(par["flip_angle(s)"])):
            phi_corr[i, :, :, :] = par["flip_angle(s)"][i] *\
                np.pi / 180 * self.fa_corr

        par["unknowns_TGV"] = 2
        par["unknowns_H1"] = 0
        par["unknowns"] = par["unknowns_TGV"]+par["unknowns_H1"]

        self.sin_phi = np.sin(phi_corr)
        self.cos_phi = np.cos(phi_corr)

        for j in range(par["unknowns"]):
            self.uk_scale.append(1)

        self.constraints.append(
            constraints(0 / self.uk_scale[0],
                        10 / self.uk_scale[0],
                        False))
        self.constraints.append(
            constraints(np.exp(-self.TR / (50)),
                        np.exp(-self.TR / (5500)),
                        True))

    def rescale(self, x):
        tmp_x = np.copy(x)
        tmp_x[0] *= self.uk_scale[0]
        tmp_x[1] = -self.TR / np.log(tmp_x[1] * self.uk_scale[1])
        return tmp_x

    def _execute_forward_2D(self, x, islice):
        print('uk_scale[1]: ', self.uk_scale[1])
        E1 = x[1, ...] * self.uk_scale[1]
        S = x[0, ...] * self.uk_scale[0] * (-E1 + 1) * \
            self.sin_phi[:, islice, ...]\
            / (-E1 * self.cos_phi[:, islice, ...] + 1)
        S[~np.isfinite(S)] = 1e-20
        S = np.array(S, dtype=DTYPE)
        return S

    def _execute_gradient_2D(self, x, islice):
        E1 = x[1, ...] * self.uk_scale[1]
        M0 = x[0, ...]
        E1[~np.isfinite(E1)] = 0
        grad_M0 = self.uk_scale[0] * (-E1 + 1) * self.sin_phi[:, islice, ...]\
            / (-E1 * self.cos_phi[:, islice, ...] + 1)
        grad_T1 = M0 * self.uk_scale[0] * self.uk_scale[1] * (-E1 + 1) *\
            self.sin_phi[:, islice, ...] * self.cos_phi[:, islice, ...] /\
            (-E1 * self.cos_phi[:, islice, ...] + 1)**2 -\
            M0 * self.uk_scale[0] * self.uk_scale[1] *\
            self.sin_phi[:, islice, ...] /\
            (-E1 * self.cos_phi[:, islice, ...] + 1)
        grad = np.array([grad_M0, grad_T1], dtype=DTYPE)
        grad[~np.isfinite(grad)] = 1e-20
        return grad

    def _execute_forward_3D(self, x):
        E1 = x[1, ...] * self.uk_scale[1]
        S = x[0, ...] * self.uk_scale[0] * (-E1 + 1) * self.sin_phi /\
            (-E1 * self.cos_phi + 1)
        S[~np.isfinite(S)] = 1e-20
        S = np.array(S, dtype=DTYPE)
        return S

    def _execute_gradient_3D(self, x):
        E1 = x[1, ...] * self.uk_scale[1]
        M0 = x[0, ...]
        E1[~np.isfinite(E1)] = 0
        grad_M0 = self.uk_scale[0] * (-E1 + 1) * self.sin_phi /\
            (-E1 * self.cos_phi + 1)
        grad_T1 = M0 * self.uk_scale[0] * self.uk_scale[1] * (-E1 + 1) *\
            self.sin_phi * self.cos_phi / (-E1 * self.cos_phi + 1)**2 -\
            M0 * self.uk_scale[0] * self.uk_scale[1] * self.sin_phi /\
            (-E1 * self.cos_phi + 1)
        grad = np.array([grad_M0, grad_T1], dtype=DTYPE)
        grad[~np.isfinite(grad)] = 1e-20
        return grad

    def plot_unknowns(self, x, dim_2D=False):
        M0 = np.abs(x[0, ...] * self.uk_scale[0])
        T1 = np.abs(-self.TR / np.log(x[1, ...] * self.uk_scale[1]))
        M0_min = M0.min()
        M0_max = M0.max()
        T1_min = T1.min()
        T1_max = T1.max()

        if dim_2D:
            if not self.figure:
                plt.ion()
                self.figure, self.ax = plt.subplots(1, 2, figsize=(12, 5))
                self.M0_plot = self.ax[0].imshow((M0))
                self.ax[0].set_title('Proton Density in a.u.')
                self.ax[0].axis('off')
                self.figure.colorbar(self.M0_plot, ax=self.ax[0])
                self.T1_plot = self.ax[1].imshow((T1))
                self.ax[1].set_title('T1 in  ms')
                self.ax[1].axis('off')
                self.figure.colorbar(self.T1_plot, ax=self.ax[1])
                self.figure.tight_layout()
                plt.draw()
                plt.pause(1e-10)
            else:
                self.M0_plot.set_data((M0))
                self.M0_plot.set_clim([M0_min, M0_max])
                self.T1_plot.set_data((T1))
                self.T1_plot.set_clim([T1_min, T1_max])
                plt.draw()
                plt.pause(1e-10)
        else:
            [z, y, x] = M0.shape
            self.ax = []
            if not self.figure:
                plt.ion()
                self.figure = plt.figure(figsize=(12, 6))
                self.figure.subplots_adjust(hspace=0, wspace=0)
                self.gs = gridspec.GridSpec(
                    2, 6,
                    width_ratios=[
                        x / (20 * z), x / z, 1, x / z, 1, x / (20 * z)],
                    height_ratios=[x / z, 1])
                self.figure.tight_layout()
                self.figure.patch.set_facecolor(plt.cm.viridis.colors[0])
                for grid in self.gs:
                    self.ax.append(plt.subplot(grid))
                    self.ax[-1].axis('off')

                self.M0_plot = self.ax[1].imshow(
                    (M0[int(self.NSlice / 2), ...]))
                self.M0_plot_cor = self.ax[7].imshow(
                    (M0[:, int(M0.shape[1] / 2), ...]))
                self.M0_plot_sag = self.ax[2].imshow(
                    np.flip((M0[:, :, int(M0.shape[-1] / 2)]).T, 1))
                self.ax[1].set_title('Proton Density in a.u.', color='white')
                self.ax[1].set_anchor('SE')
                self.ax[2].set_anchor('SW')
                self.ax[7].set_anchor('NW')
                cax = plt.subplot(self.gs[:, 0])
                cbar = self.figure.colorbar(self.M0_plot, cax=cax)
                cbar.ax.tick_params(labelsize=12, colors='white')
                cax.yaxis.set_ticks_position('left')
                for spine in cbar.ax.spines:
                    cbar.ax.spines[spine].set_color('white')
                plt.draw()
                plt.pause(1e-10)

                self.T1_plot = self.ax[3].imshow(
                    (T1[int(self.NSlice / 2), ...]))
                self.T1_plot_cor = self.ax[9].imshow(
                    (T1[:, int(T1.shape[1] / 2), ...]))
                self.T1_plot_sag = self.ax[4].imshow(
                    np.flip((T1[:, :, int(T1.shape[-1] / 2)]).T, 1))
                self.ax[3].set_title('T1 in  ms', color='white')
                self.ax[3].set_anchor('SE')
                self.ax[4].set_anchor('SW')
                self.ax[9].set_anchor('NW')
                cax = plt.subplot(self.gs[:, 5])
                cbar = self.figure.colorbar(self.T1_plot, cax=cax)
                cbar.ax.tick_params(labelsize=12, colors='white')
                for spine in cbar.ax.spines:
                    cbar.ax.spines[spine].set_color('white')
                plt.draw()
                plt.pause(1e-10)
            else:
                self.M0_plot.set_data((M0[int(self.NSlice / 2), ...]))
                self.M0_plot_cor.set_data((M0[:, int(M0.shape[1] / 2), ...]))
                self.M0_plot_sag.set_data(
                    np.flip((M0[:, :, int(M0.shape[-1] / 2)]).T, 1))
                self.M0_plot.set_clim([M0_min, M0_max])
                self.M0_plot_cor.set_clim([M0_min, M0_max])
                self.M0_plot_sag.set_clim([M0_min, M0_max])
                self.T1_plot.set_data((T1[int(self.NSlice / 2), ...]))
                self.T1_plot_cor.set_data((T1[:, int(T1.shape[1] / 2), ...]))
                self.T1_plot_sag.set_data(
                    np.flip((T1[:, :, int(T1.shape[-1] / 2)]).T, 1))
                self.T1_plot.set_clim([T1_min, T1_max])
                self.T1_plot_sag.set_clim([T1_min, T1_max])
                self.T1_plot_cor.set_clim([T1_min, T1_max])
                plt.draw()
                plt.pause(1e-10)

    def computeInitialGuess(self, *args):
        test_T1 = 1500 * np.ones(
            (self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_M0 = np.ones((self.NSlice, self.dimY, self.dimX), dtype=DTYPE)
        test_T1 = np.exp(-self.TR / (test_T1))
        x = np.array([test_M0 / self.uk_scale[0],
                      test_T1 / self.uk_scale[1]], dtype=DTYPE)
        self.guess = x
